from __future__ import division, print_function, absolute_import, unicode_literals
#*****************************************************************
#    pyGSTi 0.9:  Copyright 2015 Sandia Corporation
#    This Software is released under the GPL license detailed
#    in the file "license.txt" in the top-level pyGSTi directory
#*****************************************************************
""" Defines the EvalTree class which implements an evaluation tree. """

from . import gatestring as _gs
from ..tools import mpitools as _mpit
from .verbosityprinter import VerbosityPrinter

import numpy as _np
import time as _time #DEBUG TIMERS

class EvalTree(list):
    """
    An Evaluation Tree.  Instances of this class specify how to
      perform bulk GateSet operations.

    EvalTree instances create and store the decomposition of a list
      of gate strings into a sequence of 2-term products of smaller
      strings.  Ideally, this sequence would prescribe the way to
      obtain the entire list of gate strings, starting with just the
      single gates, using the fewest number of multiplications, but
      this optimality is not guaranteed.
    """
    def __init__(self, items=[]):
        """ Create a new, empty, evaluation tree. """
        self.gateLabels = []
        self.init_indices = []
        self.eval_order = []
        self.myFinalToParentFinalMap = None
        self.num_final_strs = 0
        self.subTrees = []
        self.parentIndexMap = None
        self.original_index_lookup = None
        self.distribution = {}
        super(EvalTree, self).__init__(items)

    def initialize(self, gateLabels, gatestring_list, numSubTreeComms=1):
        """
          Initialize an evaluation tree using a set of gate strings.
          This function must be called before using an EvalTree.

          Parameters
          ----------
          gateLabels : list of strings
              A list of all the single gate labels to
              be stored at the beginning of the tree.  This
              list must include all the gate labels contained
              in the elements of gatestring_list.

          gatestring_list : list of (tuples or GateStrings)
              A list of tuples of gate labels or GateString
              objects, specifying the gate strings that
              should be present in the evaluation tree.

          numSubTreeComms : int, optional
              The number of processor groups (communicators)
              to divide the subtrees of this EvalTree among
              when calling `distribute`.  By default, the
              communicator is not divided.

          Returns
          -------
          None
        """
        #tStart = _time.time() #DEBUG TIMER
        self.gateLabels = gateLabels
        if numSubTreeComms is not None:
            self.distribution['numSubtreeComms'] = numSubTreeComms

        if len(gatestring_list ) > 0 and isinstance(gatestring_list[0],_gs.GateString):
            gatestring_list = [gs.tup for gs in gatestring_list]

        #Evaluation dictionary:
        # keys == gate strings that have been evaluated so far
        # values == index of gate string (key) within evalTree
        evalDict = { }

        #Evaluation tree:
        # A list of tuples, where each element contains
        #  information about evaluating a particular gate string:
        #  (iLeft, iRight)
        # and the order of the elements specifies the evaluation order.
        # In particular, the gateString = evalTree[iLeft] + evalTree[iRight]
        #   so that matrix(gateString) = matrixOf(evalTree[iRight]) * matrixOf(evalTree[iLeft])
        del self[:] #clear self (a list)

        #Final Indices
        # The first len(gatestring_list) elements of the tree correspond
        # to computing the gate strings requested in gatestring_list.  Doing
        # this make later extraction much easier (views can be used), but
        # requires a non-linear order of evaluation, held in the eval_order list.
        self.eval_order = []

        #initialize self as a list of Nones
        self.num_final_strs = len(gatestring_list)
        self[:] = [None]*self.num_final_strs

        #Single gate (or zero-gate) computations are assumed to be atomic, and be computed independently.
        #  These labels serve as the initial values, and each gate string is assumed to be a tuple of
        #  gate labels.
        self.init_indices = [] #indices to put initial zero & single gate results
        for gateLabel in self.gateLabels:
            tup = () if gateLabel == "" else (gateLabel,) #special case of empty label == no gate
            if tup in gatestring_list:
                indx = gatestring_list.index(tup) 
                self[indx] = (None,None) #iLeft = iRight = None for always-evaluated zero string
            else:
                indx = len(self)
                self.append( (None,None) ) #iLeft = iRight = None for always-evaluated zero string
            self.init_indices.append( indx )
            evalDict[ tup ] = indx

        #print("DB: initial eval dict = ",evalDict)

        #Process gatestrings in order of length, so that we always place short strings
        # in the right place (otherwise assert stmt below can fail)
        indices_sorted_by_gatestring_len = \
            sorted(list(range(len(gatestring_list))),
                   key=lambda i: len(gatestring_list[i]))

        #avgBiteSize = 0
        #useCounts = {}
        #OLD (sequential): for (k,gateString) in enumerate(gatestring_list):
        for k in indices_sorted_by_gatestring_len:
            gateString = gatestring_list[k]
            L = len(gateString)
            if L == 0:
                iEmptyStr = evalDict.get( (), None)
                assert(iEmptyStr is not None) # duplicate () final strs require
                if k != iEmptyStr:
                    assert(self[k] is None)       # the empty string to be included in the tree too!
                    self[k] = (iEmptyStr, iEmptyStr) # compute the duplicate () using by
                    self.eval_order.append(k)        #  multiplying by the empty string.

            start = 0; bite = 1
            #nBites = 0
            #print("\nDB: string = ",gateString, "(len=%d)" % len(gateString))

            while start < L:

                #Take a bite out of gateString, starting at `start` that is in evalDict
                for b in range(L-start,0,-1):
                    if gateString[start:start+b] in evalDict:
                        bite = b; break
                else: assert(False) #Logic error - loop above should always exit when b == 1

                #iInFinal = k if bool(start + bite == L) else -1
                bFinal = bool(start + bite == L)
                #print("DB: start=",start,": found ",gateString[start:start+bite],
                #      " (len=%d) in evalDict" % bite, "(final=%s)" % bFinal)

                if start == 0: #first in-evalDict bite - no need to add anything to self yet
                    iCur = evalDict[ gateString[0:bite] ]
                    #print("DB: taking bite: ", gateString[0:bite], "indx = ",iCur)
                    if bFinal:
                        if iCur != k:  #then we have a duplicate final gate string
                            iEmptyStr = evalDict.get( (), None)
                            assert(iEmptyStr is not None) # duplicate final strs require
                                      # the empty string to be included in the tree too!
                            assert(self[k] is None) #make sure we haven't put anything here yet
                            self[k] = (iCur, iEmptyStr) # compute the duplicate using by                             
                            self.eval_order.append(k)   #  multiplying by the empty string.
                else:
                    # add (iCur, iBite)
                    assert(gateString[0:start+bite] not in evalDict)
                    iBite = evalDict[ gateString[start:start+bite] ]
                    if bFinal: #place (iCur, iBite) at location k
                        iNew = k
                        evalDict[ gateString[0:start+bite] ] = iNew
                        assert(self[iNew] is None) #make sure we haven't put anything here yet
                        self[k] = (iCur, iBite)
                    else:
                        iNew = len(self)
                        evalDict[ gateString[0:start+bite] ] = iNew
                        self.append( (iCur,iBite) )

                    #print("DB: add %s (index %d)" % (str(gateString[0:start+bite]),iNew))
                    self.eval_order.append(iNew)
                    iCur = iNew
                start += bite
                #nBites += 1

            #if nBites > 0: avgBiteSize += L / float(nBites)
            assert(k in self.eval_order or k in self.init_indices)
        
        #avgBiteSize /= float(len(gatestring_list))
        #print "DEBUG: Avg bite size = ",avgBiteSize

        #see if there are superfluous tree nodes: those with iFinal == -1 and
        self.myFinalToParentFinalMap = None #this tree has no "children",
        self.parentIndexMap = None          # i.e. has not been created by a 'split'
        self.original_index_lookup = None
        self.subTrees = [] #no subtrees yet
        assert(self.generate_gatestring_list() == gatestring_list)
        assert(None not in gatestring_list)


    def copy(self):
        """ Create a copy of this evaluation tree. """
        newTree = EvalTree(self[:])
        newTree.gateLabels = self.gateLabels[:]
        newTree.init_indices = self.init_indices[:]
        newTree.eval_order = self.eval_order[:]
        newTree.num_final_strs = self.num_final_strs
        newTree.myFinalToParentFinalMap = self.myFinalToParentFinalMap
        newTree.parentIndexMap = self.parentIndexMap[:] \
            if (self.parentIndexMap is not None) else None
        newTree.subTrees = [ st.copy() for st in self.subTrees ]
        newTree.original_index_lookup = self.original_index_lookup[:] \
            if (self.original_index_lookup is not None) else None
        return newTree

    def get_init_labels(self):
        """ Return a tuple of the gate labels (strings)
            which form the beginning of the tree.
        """
        return tuple(self.gateLabels)

    def get_init_indices(self):
        """ Return a tuple of the indices corresponding
             to the initial gate labels (strings)
             which form the beginning of the tree.
        """
        return tuple(self.init_indices)

    def get_evaluation_order(self):
        """ Return a list of indices specifying the
             order in which elements of this EvalTree
             should be visited when doing a computation
             (after computing the initial indices).
        """
        return self.eval_order


    def final_view(self, a, axis=None):
        """ 
        Returns a view of array `a` restricting it to only the
        *final* results computed by this tree (not the intermediate
        results).

        Parameters
        ----------
        a : ndarray
            An array of results computed using this EvalTree, 
            such that the `axis`-th dimension equals the full
            length of the tree.  The other dimensions of `a` are
            unrestricted.

        axis : int, optional
            Specified the axis along which the selection of the
            final elements is performed. If None, than this 
            selection if performed on flattened `a`.

        Returns
        -------
        ndarray
            Of the same shape as `a`, except for along the 
            specified axis, whose dimension has been reduced
            to filter out the intermediate (non-final) results.
        """
        if axis is None:
            return a[0:self.num_final_strings()]
        else:
            sl = [slice(None)] * a.ndim
            sl[axis] = slice(0,self.num_final_strings())
            ret = a[sl]
            assert(ret.base is a) #check that what is returned is a view
            assert(ret.size == 0 or _np.may_share_memory(ret,a))
            return ret


    def final_slice(self, parent_tree):
        """
        Return a slice that identifies the segment of `parent_tree`'s
        final values that correspond to this tree's final values.

        Note that if `parent_tree` is not None, this tree must be a
        sub-tree of it (which means `parent_tree` is split).

        Parameters
        ----------
        parent_tree : EvalTree
            This tree's parent tree.  If the parent tree is None
            or isn't split (which sometimes means that this 
            "sub-tree" is the same as the parent tree - see 
            `get_sub_trees`), then a slice for the entire set of
            final values is returned, which is appropriate in this
            case.

        Returns
        -------
        slice
        """
        if (self.myFinalToParentFinalMap is not None) and \
                parent_tree.is_split():
            return self.myFinalToParentFinalMap
        else:
            return slice(0,self.num_final_strings())
        

    def num_final_strings(self):
        """
        Returns the integer number of "final" gate strings, equal
          to the length of the gatestring_list passed to initialize.
        """
        return self.num_final_strs


    def generate_gatestring_list(self, permute=True):
        """
        Generate a list of the final gate strings this tree evaluates.

        This method essentially "runs" the tree and follows its
          prescription for sequentailly building up longer strings
          from shorter ones.  When permute == True, the resulting list
          should be the same as the one passed to initialize(...), and
          so this method may be used as a consistency check.

        Parameters
        ----------
        permute : bool, optional
           Whether to permute the returned list of strings into the 
           same order as the original list passed to initialize(...).
           When False, the computed order of the gate strings is 
           given, which is matches the order of the results from calls
           to `GateSet` bulk operations.  Non-trivial permutation
           occurs only when the tree is split (in order to keep 
           each sub-tree result a contiguous slice within the parent
           result).

        Returns
        -------
        list of gate-label-tuples
            A list of the gate strings evaluated by this tree, each
            specified as a tuple of gate labels.
        """
        gateStrings = [None]*len(self)

        #Set "initial" (single- or zero- gate) strings
        for i,gateLabel in zip(self.get_init_indices(), self.get_init_labels()):
            if gateLabel == "": gateStrings[i] = () #special case of empty label
            else: gateStrings[i] = (gateLabel,)

        #Build rest of strings
        for i in self.get_evaluation_order():
            iLeft, iRight = self[i]
            gateStrings[i] = gateStrings[iLeft] + gateStrings[iRight]
            
        #Permute to get final list:
        nFinal = self.num_final_strings()
        if self.original_index_lookup is not None and permute == True:
            finalGateStrings = [None]*nFinal
            for iorig,icur in self.original_index_lookup.items():
                if iorig < nFinal: finalGateStrings[iorig] = gateStrings[icur]
            assert(None not in finalGateStrings)
            return finalGateStrings
        else:
            assert(None not in gateStrings[0:nFinal])
            return gateStrings[0:nFinal]


    def permute_original_to_computation(self, a, axis=0):
        """
        Permute an array's elements using mapping from the "original"
        gate string ordering to the "computation" ordering.
        
        This function converts arrays with elements corresponding
        to gate strings in the "original" ordering (i.e. the 
        ordering in the list passed to `initialize(...)`) to the
        ordering used in tree computation (i.e. by a `GateSet`'s
        bulk computation routines).

        Paramters
        ---------
        a : numpy array
           The array whose permuted elements are returned.

        axis : int, optional
           The axis to permute.  By default, the first dimension is used.

        Returns
        -------
        numpy array
        """
        assert(a.shape[axis] == self.num_final_strings())
        nFinal = self.num_final_strings()
        ret = a.copy()

        def mkindx(i):
            mi = [slice(None)]*a.ndim; mi[axis] = i
            return mi

        if self.original_index_lookup is not None:
            for iorig,icur in self.original_index_lookup.items():                
                if iorig < nFinal: 
                    ret[mkindx(icur)] = a[mkindx(iorig)]

        return ret


    def permute_computation_to_original(self, a, axis=0):
        """
        Permute an array's elements using mapping from the "computation"
        gate string ordering to the "original" ordering.
        
        This function converts arrays with elements corresponding
        to gate strings in the ordering used in tree computation 
        (i.e. the ordering returned by `GateSet`'s bulk computation routines)
        to the "original" ordering (i.e. the ordering of the gate string list
        passed to `initialize(...)`).

        Paramters
        ---------
        a : numpy array
           The array whose permuted elements are returned.

        axis : int, optional
           The axis to permute.  By default, the first dimension is used.

        Returns
        -------
        numpy array
        """
        assert(a.shape[axis] == self.num_final_strings())
        nFinal = self.num_final_strings()
        ret = a.copy()

        def mkindx(i):
            mi = [slice(None)]*a.ndim; mi[axis] = i
            return mi

        if self.original_index_lookup is not None:
            for iorig,icur in self.original_index_lookup.items():                
                if iorig < nFinal: 
                    ret[mkindx(iorig)] = a[mkindx(icur)]

        return ret
        

    def distribute(self, comm, verbosity=0):
        """
        Distributes this tree's sub-trees across multiple processors.
        
        This function checks how many processors are present in
        `comm` and divides this tree's subtrees into groups according to
        the number of subtree comms provided as an argument to 
        `initialize`.  Note that this does *not* always divide the subtrees
        among the processors as much as possible, as this is not always what is
        desired (computing several subtrees serially using multiple
        processors simultaneously can be faster, due to better balancing, than
        computing all the subtrees simultaneously using smaller processor groups
        for each).  

        For example, if the number of subtree comms given to 
        `initialize` == 1, then all the subtrees are assigned to the one and
        only processor group containing all the processors of `comm`.  If the
        number of subtree comms == 2 and there are 4 subtrees and 8 processors,
        then `comm` is divided into 2 groups of 4 processors each, and two 
        subtrees are assigned to each processor group.

        Parameters
        ----------
        comm : mpi4py.MPI.Comm
            When not None, an MPI communicator for distributing subtrees
            across processor groups

        verbosity : int, optional
            How much detail to send to stdout.
            
        Returns
        -------
        mySubtreeIndices : list
            A list of integer indices specifying which subtrees this
            processor is responsible for.
        subTreeOwners : dict
            A dictionary whose keys are integer subtree indices and 
            whose values are processor ranks, which indicates which 
            processor is responsible for communicating the final 
            results of each subtree.
        mySubComm : mpi4py.MPI.Comm or None
            The communicator for the processor group that is responsible
            for computing the same `mySubTreeIndices` list.  This 
            communicator is used for further processor division (e.g.
            for parallelization across derivative columns).
        """
        # split tree into local subtrees, each which contains one/group of
        # processors (group can then parallelize derivative calcs over
        # gate set parameters) 

        rank = 0 if (comm is None) else comm.Get_rank()
        nprocs = 1 if (comm is None) else comm.Get_size()
        nSubtreeComms = self.distribution.get('numSubtreeComms',1)
        nSubtrees = len(self.get_sub_trees())

        assert(nSubtreeComms <= nprocs) # => len(mySubCommIndices) == 1
        mySubCommIndices, subCommOwners, mySubComm = \
            _mpit.distribute_indices(list(range(nSubtreeComms)), comm)
        assert(len(mySubCommIndices) == 1)
        mySubCommIndex = mySubCommIndices[0]

        assert(nSubtreeComms <= nSubtrees) # don't allow more comms than trees
        mySubtreeIndices, subTreeOwners = _mpit.distribute_indices_base(
            list(range(nSubtrees)), nSubtreeComms, mySubCommIndex)

        # subTreeOwners contains index of owner subComm, but we really want
        #  the owning processor, i.e. the owner of the subComm
        subTreeOwners = { iSubTree: subCommOwners[subTreeOwners[iSubTree]]
                          for iSubTree in subTreeOwners }

        printer = VerbosityPrinter.build_printer(verbosity, comm)
        printer.log("*** Distributing %d subtrees into %d sub-comms (%s processors) ***"% \
                        (nSubtrees, nSubtreeComms, nprocs))

        return mySubtreeIndices, subTreeOwners, mySubComm


    def get_min_tree_size(self):
        """
        Returns the minimum sub tree size required to compute each
        of the tree entries individually.  This minimum size is the
        smallest "maxSubTreeSize" that can be passed to split(),
        as any smaller value will result in at least one entry being
        uncomputable.
        """
        singleItemTreeSetList = self._createSingleItemTrees()
        return max(list(map(len,singleItemTreeSetList)))


    def split(self, maxSubTreeSize=None, numSubTrees=None, verbosity=0):
        """
        Split this tree into sub-trees in order to reduce the
          maximum size of any tree (useful for limiting memory consumption
          or for using multiple cores).  Must specify either maxSubTreeSize
          or numSubTrees.

        Parameters
        ----------
        maxSubTreeSize : int, optional
            The maximum size (i.e. list length) of each sub-tree.  If the
            original tree is smaller than this size, no splitting will occur.
            If None, then there is no limit.

        numSubTrees : int, optional
            The maximum size (i.e. list length) of each sub-tree.  If the
            original tree is smaller than this size, no splitting will occur.

        verbosity : int, optional
            How much detail to send to stdout.

        Returns
        -------
        None
        """
        #dbList = self.generate_gatestring_list()
        tm = _time.time()
        printer = VerbosityPrinter.build_printer(verbosity)

        if (maxSubTreeSize is None and numSubTrees is None) or \
           (maxSubTreeSize is not None and numSubTrees is not None):
            raise ValueError("Specify *either* maxSubTreeSize or numSubTrees")
        if numSubTrees is not None and numSubTrees <= 0:
            raise ValueError("EvalTree split() error: numSubTrees must be > 0!")

        #Don't split at all if it's unnecessary
        if maxSubTreeSize is None or len(self) < maxSubTreeSize:
            if numSubTrees is None or numSubTrees == 1: return

        self.subTrees = []
        subTreesFinalList = [None]*self.num_final_strings()
        printer.log("EvalTree.split done initial prep in %.0fs" %
                    (_time.time()-tm)); tm = _time.time()

        #First pass - identify which indices go in which subtree
        #   Part 1: create disjoint set of subtrees generated by single items
        singleItemTreeSetList = self._createSingleItemTrees()
          #each element represents a subtree, and
          # is a set of the indices owned by that subtree
        nSingleItemTrees = len(singleItemTreeSetList)

        printer.log("EvalTree.split created singles in %.0fs" %
                    (_time.time()-tm)); tm = _time.time()

        #   Part 2: determine whether we need to split/merge "single" trees
        if numSubTrees is not None:

            #Merges: find the best merges to perform if any are required
            if nSingleItemTrees > numSubTrees:

                #Find trees that have least intersection to begin:
                # The goal is to find a set of single-item trees such that
                # none of them intersect much with any other of them.
                #
                # Algorithm: 
                #   - start with a set of the one tree that has least
                #       intersection with any other tree.
                #   - iteratively add the tree that has the least intersection
                #       with the trees in the existing set
                iStartingTrees = []
                start_select_method = "fast"

                if start_select_method == "best":
                    availableIndices = list(range(nSingleItemTrees))
                    i_min,_ = min( enumerate(  #index of a tree in the minimal intersection
                            ( min((len(s1.intersection(s2)) for s2 in singleItemTreeSetList[i+1:]))
                              for i,s1 in enumerate(singleItemTreeSetList[:-1]) )),
                                   key=lambda x: x[1]) #argmin using generators (np.argmin doesn't work)
                    iStartingTrees.append(i_min)
                    startingTreeEls = singleItemTreeSetList[i_min].copy()
                    del availableIndices[i_min]
                                        
                    while len(iStartingTrees) < numSubTrees:
                        ii_min,_ = min( enumerate(
                            ( len(startingTreeEls.intersection(singleItemTreeSetList[i])) 
                              for i in availableIndices )), key=lambda x: x[1]) #argmin
                        i_min = availableIndices[ii_min]
                        iStartingTrees.append(i_min)
                        startingTreeEls.update( singleItemTreeSetList[i_min] )
                        del availableIndices[ii_min]
                    
                    printer.log("EvalTree.split found starting trees in %.0fs" %
                                (_time.time()-tm)); tm = _time.time()

                elif start_select_method == "fast":
                    def get_start_indices(maxIntersect):
                        starting = [0] #always start with 0th tree
                        startingSet = singleItemTreeSetList[0].copy() 
                        for i,s in enumerate(singleItemTreeSetList[1:],start=1):
                            if len(startingSet.intersection(s)) <= maxIntersect:
                                starting.append(i)
                                startingSet.update(s)
                        return starting,startingSet

                    left,right = 0, max(map(len,singleItemTreeSetList))
                    while left < right:
                        mid = (left+right) // 2
                        iStartingTrees,startingTreeEls = get_start_indices(mid)
                        nStartingTrees = len(iStartingTrees)
                        if nStartingTrees < numSubTrees:
                            left = mid + 1
                        elif nStartingTrees > numSubTrees:
                            right = mid
                        else: break # nStartingTrees == numSubTrees!

                    if len(iStartingTrees) < numSubTrees:
                        iStartingTrees,startingTreeEls = get_start_indices(mid+1)
                    if len(iStartingTrees) > numSubTrees:
                        iStartingTrees = iStartingTrees[0:numSubTrees]
                        startingTreeEls = set()
                        for i in iStartingTrees:
                            startingTreeEls.update(singleItemTreeSetList[i])
                                        
                    printer.log("EvalTree.split fast-found starting trees in %.0fs" %
                                (_time.time()-tm)); tm = _time.time()

                else:
                    raise ValueError("Invalid start select method: %s" % start_select_method)


                #Merge all the non-starting trees into the starting trees
                # so that we're left with the desired number of trees
                subTreeSetList = [singleItemTreeSetList[i] for i in iStartingTrees]
                assert(len(subTreeSetList) == numSubTrees)

                indicesLeft = list(range(nSingleItemTrees))
                for i in iStartingTrees:
                    del indicesLeft[indicesLeft.index(i)]

                printer.log("EvalTree.split deleted initial indices in %.0fs" %
                            (_time.time()-tm)); tm = _time.time()
                merge_method = "fast"

                if merge_method == "best":
                    while len(indicesLeft) > 0:
                        iToMergeInto,_ = min(enumerate(map(len,subTreeSetList)), 
                                             key=lambda x: x[1]) #argmin
                        setToMergeInto = subTreeSetList[iToMergeInto]
                        #intersectionSizes = [ len(setToMergeInto.intersection(
                        #            singleItemTreeSetList[i])) for i in indicesLeft ]
                        #iMaxIntsct = _np.argmax(intersectionSizes)
                        iMaxIntsct,_ = max( enumerate( ( len(setToMergeInto.intersection(
                                            singleItemTreeSetList[i])) for i in indicesLeft )),
                                          key=lambda x: x[1]) #argmax
                        setToMerge = singleItemTreeSetList[indicesLeft[iMaxIntsct]]
                        subTreeSetList[iToMergeInto] = \
                              subTreeSetList[iToMergeInto].union(setToMerge)
                        del indicesLeft[iMaxIntsct]
                        
                elif merge_method == "fast":
                    most_at_once = 10
                    desiredLength = int(len(self) / numSubTrees)
                    while len(indicesLeft) > 0:
                        iToMergeInto,_ = min(enumerate(map(len,subTreeSetList)), 
                                             key=lambda x: x[1]) #argmin
                        setToMergeInto = subTreeSetList[iToMergeInto]
                        intersectionSizes = sorted( [ (ii,len(setToMergeInto.intersection(
                                        singleItemTreeSetList[i]))) for ii,i in enumerate(indicesLeft) ],
                                                    key=lambda x: x[1], reverse=True)
                        toDelete = []
                        for i in range(min(most_at_once,len(indicesLeft))):
                            #if len(subTreeSetList[iToMergeInto]) >= desiredLength: break
                            iMaxIntsct,_ = intersectionSizes[i]
                            setToMerge = singleItemTreeSetList[indicesLeft[iMaxIntsct]]
                            subTreeSetList[iToMergeInto].update(setToMerge)
                            toDelete.append(iMaxIntsct)
                        for i in sorted(toDelete,reverse=True):
                            del indicesLeft[i]

                else:
                    raise ValueError("Invalid merge method: %s" % merge_method)


                assert(len(subTreeSetList) == numSubTrees)
                printer.log("EvalTree.split merged trees in %.0fs" %
                            (_time.time()-tm)); tm = _time.time()

            #Splits (more subtrees desired than there are single item trees!)
            else:
                #Splits: find the best splits to perform
                #TODO: how to split a tree intelligently -- for now, just do
                # trivial splits by making empty trees.
                subTreeSetList = singleItemTreeSetList[:]
                nSplitsNeeded = numSubTrees - nSingleItemTrees
                while nSplitsNeeded > 0:
                    # LATER...
                    # for iSubTree,subTreeSet in enumerate(subTreeSetList):
                    subTreeSetList.append( [] ) # create empty subtree
                    nSplitsNeeded -= 1

        else:
            assert(maxSubTreeSize is not None)
            subTreeSetList = []

            #Merges: find the best merges to perform if any are allowed given
            # the maximum tree size
            for singleItemTreeSet in singleItemTreeSetList:
                if len(singleItemTreeSet) > maxSubTreeSize:
                    raise ValueError("Max. sub tree size (%d) is too low (<%d)!"
                                   % (maxSubTreeSize, self.get_min_tree_size()))

                #See if we should merge this single-item-generated tree with
                # another one or make it a new subtree.
                newTreeSize = len(singleItemTreeSet)
                maxIntersectSize = None; iMaxIntersectSize = None
                for k,existingSubTreeSet in enumerate(subTreeSetList):
                    mergedSize = len(existingSubTreeSet) + newTreeSize
                    if mergedSize <= maxSubTreeSize:
                        intersectionSize = \
                            len(singleItemTreeSet.intersection(existingSubTreeSet))
                        if maxIntersectSize is None or \
                                maxIntersectSize < intersectionSize:
                            maxIntersectSize = intersectionSize
                            iMaxIntersectSize = k

                if iMaxIntersectSize is not None:
                    # then we merge the new tree with this existing set
                    subTreeSetList[iMaxIntersectSize] = \
                      subTreeSetList[iMaxIntersectSize].union(singleItemTreeSet)
                else: # we create a new subtree
                    subTreeSetList.append( singleItemTreeSet )

        #TODO: improve tree efficiency via better splitting?
        #print "DEBUG TREE SPLITTING:"
        #for k,dbTreeSet in enumerate(subTreeSetList):
        #    print "Tree %d (size %d): " % (k,len(dbTreeSet)), [ len(dbTreeSet.intersection(x)) for kk,x in enumerate(subTreeSetList) if kk != k ]
        #cnts = [0]*len(self)
        #for k,dbTreeSet in enumerate(subTreeSetList):
        #    for i in dbTreeSet:
        #        cnts[i] += 1
        #sorted_cnts = sorted( list(enumerate(cnts)), key=lambda x: x[1], reverse=True)
        #print "Top index : cnts"
        #for ii,(i,cnt) in enumerate(sorted_cnts):
        #    print ii,":", i,", ",cnt
        #raise ValueError("STOP")
                    
        #bDebug = False
        #if bDebug: print("Parent nFinal = ",self.num_final_strings(), " len=",len(self))
        printer.log("EvalTree.split done first pass in %.0fs" %
                    (_time.time()-tm)); tm = _time.time()

        #Second pass - create subtrees from index sets
        need_to_compute = _np.zeros( len(self), 'bool' )
        need_to_compute[0:self.num_final_strings()] = True

          #First, reorder the parent tree's elements so that the final
          # elements of the subtrees map to contiguous slices of the
          # parent tree's final elements.
        parentIndexRevPerm = [] # parentIndexRevPerm[newIndex] = currentIndex (i.e. oldIndex)
        subTreeIndicesList = []
        numFinalList = []
        for iSubTree,subTreeSet in enumerate(subTreeSetList):
            subTreeIndices = list(subTreeSet)
            #if bDebug: print("SUBTREE0: %s (len=%d)" % (str(subTreeIndices),len(subTreeIndices)))
            #if bDebug: print("  NEED: %s" % ",".join([ "1" if b else "0" for b in need_to_compute]))
            subTreeIndices.sort() # order subtree gatestrings (really just their indices) so
                                  # that all "final" strings come first.
            #Compute # of "final" strings in this subtree (count # of indices < num_final_strs)
            subTreeNumFinal = _np.sum(_np.array(subTreeIndices) < self.num_final_strings())

            #Swap the indices of "final" strings that have already been computed past the end
            # of the "final strings" region of the subtree's list (i.e. the subtree itself).
            already_computed = _np.logical_not( need_to_compute[ subTreeIndices[0:subTreeNumFinal] ] )
            already_computed_inds = _np.nonzero(already_computed)[0] # (sorted ascending)
            #if bDebug: print("SUBTREE1: %s (nFinal=%d - %d)" % (str(subTreeIndices),
            #                                         subTreeNumFinal, len(already_computed_inds)))
            #if bDebug: print("  - already computed = ", [subTreeIndices[i] for i in already_computed_inds])

            iFirstNonFinal = subTreeNumFinal
            for k in already_computed_inds:
                if k >= iFirstNonFinal: continue #already a non-final el
                elif k == iFirstNonFinal-1: #index is last "final" el - just shift boundary
                    iFirstNonFinal -= 1 #now index is "non-final"
                else: # k < iFirstNonFinal-1, so find a desired "final" el at boundary to swap it with
                    iLastFinal = iFirstNonFinal-1
                    while iLastFinal > k and (iLastFinal in already_computed_inds):
                        iLastFinal -= 1 #the element at iLastFinal happens to be one that we wanted to be non-final, so remove it
                    if iLastFinal != k:
                        subTreeIndices[iLastFinal],subTreeIndices[k] = \
                            subTreeIndices[k],subTreeIndices[iLastFinal] #Swap k <-> iLastFinal
                    iFirstNonFinal = iLastFinal # move boundary to make k's new location non-final

            subTreeNumFinal = iFirstNonFinal # the final <-> non-final boundary
            parentIndexRevPerm.extend( subTreeIndices[0:subTreeNumFinal] )
            subTreeIndicesList.append( subTreeIndices )
            numFinalList.append( subTreeNumFinal )
            need_to_compute[ subTreeIndices[0:subTreeNumFinal] ] = False
            #if bDebug: print("FINAL SUBTREE: %s (nFinal=%d)" % (str(subTreeIndices),subTreeNumFinal))
                    
        #Permute parent tree indices according to parentIndexPerm
        assert(len(parentIndexRevPerm) == self.num_final_strings())
        parentIndexRevPerm.extend( list(range(self.num_final_strings(), len(self))) ) 
          #don't permute non-final indices (no need)
        
        #Create forward permutation map: currentIndex -> newIndex
        parentIndexPerm = [ None ] * len(parentIndexRevPerm)
        for inew,icur in enumerate(parentIndexRevPerm):
            parentIndexPerm[icur] = inew
        assert( None not in parentIndexPerm) #all indices should be mapped somewhere!
        assert( self.original_index_lookup is None )
        self.original_index_lookup = { icur: inew for inew,icur in enumerate(parentIndexRevPerm) }

        #if bDebug: print("PERM REV MAP = ", parentIndexRevPerm)
        #if bDebug: print("PERM MAP = ", parentIndexPerm)

        #Permute parent indices
        self.init_indices = [ parentIndexPerm[iCur] for iCur in self.init_indices ]
        self.eval_order = [ parentIndexPerm[iCur] for iCur in self.eval_order ]
        self[:] = [ (parentIndexPerm[self[iCur][0]] if (self[iCur][0] is not None) else None,
                     parentIndexPerm[self[iCur][1]] if (self[iCur][1] is not None) else None)
                    for iCur in parentIndexRevPerm ]
        assert(self.myFinalToParentFinalMap is None)
        assert(self.parentIndexMap is None)


        #Update indices in subtree lists
        newList = []; finalSlices = []; sStart = 0
        for subTreeIndices,numFinal in zip(subTreeIndicesList,numFinalList):
            newSubTreeIndices = [ parentIndexPerm[i] for i in subTreeIndices ] 
            assert(newSubTreeIndices[0:numFinal] == list(range(sStart,sStart+numFinal)))
              #final elements should be a sequential slice of parent indices
            finalSlices.append( slice(sStart,sStart+numFinal) )
            newList.append(newSubTreeIndices)
            sStart += numFinal #increment slice start position
        subTreeIndicesList = newList

        for iSubTree,(subTreeIndices,subTreeNumFinal,slc) \
                in enumerate(zip(subTreeIndicesList,numFinalList,finalSlices)):

            fullEvalOrder = self.init_indices + self.eval_order #list concatenation
            subTreeEvalOrder = sorted(range(len(subTreeIndices)),
                                      key=lambda j: fullEvalOrder.index(subTreeIndices[j]))
            mapIndxToSubTreeIndx = { k: ik for ik,k in enumerate(subTreeIndices) }

            subTree = EvalTree()
            subTree.myFinalToParentFinalMap = slc
            subTree.num_final_strs = subTreeNumFinal
            subTree[:] = [None]*len(subTreeIndices)

            for ik in subTreeEvalOrder:
                k = subTreeIndices[ik] #original tree index
                (oLeft,oRight) = self[k] #original tree indices

                if (oLeft is None) and (oRight is None):
                    iLeft = iRight = None
                    #assert(len(subTree.gateLabels) == len(subTree)) #make sure all gatelabel items come first
                    subTree.gateLabels.append( self.gateLabels[ 
                            self.init_indices.index(k)] )
                    subTree.init_indices.append(ik)
                else:
                    iLeft  = mapIndxToSubTreeIndx[ oLeft ]
                    iRight = mapIndxToSubTreeIndx[ oRight ]
                    subTree.eval_order.append(ik)

                assert(subTree[ik] is None)
                subTree[ik] = (iLeft,iRight)

                #if ik < subTreeNumFinal:
                #    assert(k < self.num_final_strings()) # it should be a final element in parent too!
                #    subTree.myFinalToParentFinalMap[ik] = k

            subTree.parentIndexMap = subTreeIndices #parent index of each subtree index
            self.subTrees.append( subTree )
            #if bDebug: print("NEW SUBTREE: indices=%s, len=%d, nFinal=%d"  %
            #                  (subTree.parentIndexMap, len(subTree), subTree.num_final_strings()))

        #dbList2 = self.generate_gatestring_list()
        #if bDebug: print("DBLIST = ",dbList)
        #if bDebug: print("DBLIST2 = ",dbList2)
        #assert(dbList == dbList2)

        printer.log("EvalTree.split done second pass in %.0fs" %
                    (_time.time()-tm)); tm = _time.time()

        return


    def is_split(self):
        """ Returns boolean indicating whether tree is split into sub-trees or not. """
        return len(self.subTrees) > 0


    def get_sub_trees(self):
        """
        Returns a list of all the sub-trees (also EvalTree instances) of
          this tree.  If this tree is not split, returns a single-element
          list containing just the tree.
        """
        if self.is_split():
            return self.subTrees
        else:
            return [self] #return self as the only "subTree" when not split

    def _walkSubTree(self,indx,out):
        if indx not in out: out.append(indx)
        (iLeft,iRight) = self[indx]
        if iLeft is not None: self._walkSubTree(iLeft,out)
        if iRight is not None: self._walkSubTree(iRight,out)

    def _createSingleItemTrees(self):
        #  Create disjoint set of subtrees generated by single items
        need_to_compute = _np.zeros( len(self), 'bool' )
        need_to_compute[0:self.num_final_strings()] = True

        singleItemTreeSetList = [] #each element represents a subtree, and
                            # is a set of the indices owned by that subtree
        for i in reversed(range(self.num_final_strings())):
            if not need_to_compute[i]: continue # move to the last element
              #of evalTree that needs to be computed (i.e. is not in a subTree)

            subTreeIndices = [] # create subtree for uncomputed item
            self._walkSubTree(i,subTreeIndices)
            newTreeSet = set(subTreeIndices)
            for k in subTreeIndices:
                need_to_compute[k] = False #mark all the elements of
                                           #the new tree as computed

            # Add this single-item-generated tree as a new subtree. Later
            #  we merge and/or split these trees based on constraints.
            singleItemTreeSetList.append( newTreeSet )
        return singleItemTreeSetList


    def print_analysis(self):
        """
        Print a brief analysis of this tree. Used for
        debugging and assessing tree quality.
        """

        #Analyze tree
        if not self.is_split():
            print("Size of evalTree = %d" % len(self))
            print("Size of gatestring_list = %d" % self.num_final_strings())

            #TODO: maybe revive this later if it ever becomes useful again.
            # Currently left un-upgraded after evaltree was changed to hold
            # all final indices at its beginning. (need to iterate in eval_order
            #  not not just over enumerate(self) as is done below).
            #lastOccurrance = [-1] * len(self)
            #nRefs = [-1] * len(self)
            #for i,tup in enumerate(self):
            #    iLeft,iRight = tup
            #
            #    if iLeft is not None:
            #        nRefs[iLeft] += 1
            #        lastOccurrance[iLeft] = i
            #
            #    if iRight is not None:
            #        nRefs[iRight] += 1
            #        lastOccurrance[iRight] = i
            #
            ##print "iTree  nRefs lastOcc  iFinal  inUse"
            #maxInUse = nInUse = 0
            #for i,tup in enumerate(self):
            #    nInUse += 1
            #    for j in range(i):
            #        if lastOccurrance[j] == i and self[j][2] == -1: nInUse -= 1
            #    maxInUse = max(maxInUse,nInUse)
            #    #print "%d  %d  %d  %d  %d" % (i, nRefs[i], lastOccurrance[i], tup[2], nInUse)
            #print("Max in use at once = (smallest tree size for mem) = %d" % maxInUse)

        else: #tree is split
            print("Size of original tree = %d" % len(self))
            print("Size of original gatestring_list = %d" % self.num_final_strings())
            print("Tree is split into %d sub-trees" % len(self.subTrees))
            print("Sub-tree lengths = ", list(map(len,self.subTrees)), " (Sum = %d)" % sum(map(len,self.subTrees)))
            for i,t in enumerate(self.subTrees):
                print(">> sub-tree %d: " % i)
                t.print_analysis()


    def get_analysis_plot_infos(self):
        """
        Returns debug plot information useful for
        assessing the quality of a tree. This
        function is not guaranteed to work.
        """

        analysis = {}
        firstIndxSeen = list(range(len(self)))
        lastIndxSeen = list(range(len(self)))
        subTreeSize = [-1]*len(self)

        xs = []; ys = []
        for i in range(len(self)):
            subTree = []
            self._walkSubTree(i,subTree)
            subTreeSize[i] = len(subTree)
            ys.extend( [i]*len(subTree) + [None] )
            xs.extend( list(sorted(subTree) + [None]) )

            for k,t in enumerate(self):
                iLeft,iRight = t
                if i in (iLeft,iRight):
                    lastIndxSeen[i] = k

        analysis['SubtreeUsagePlot'] = { 'xs': xs, 'ys': ys, 'title': "Indices used by the subtree rooted at each index",
                                                'xlabel': "Indices used", 'ylabel': 'Subtree root index' }
        analysis['SubtreeSizePlot'] = { 'xs': list(range(len(self))), 'ys': subTreeSize, 'title': "Size of subtree rooted at each index",
                                                'xlabel': "Subtree root index", 'ylabel': 'Subtree size' }

        xs = [];  ys = []
        for i,rng in enumerate(zip(firstIndxSeen,lastIndxSeen)):
            ys.extend( [i,i,None] )
            xs.extend( [rng[0],rng[1],None] )
        analysis['IndexUsageIntervalsPlot'] = { 'xs': xs, 'ys': ys, 'title': "Usage Intervals of each index",
                                                'xlabel': "Index Interval", 'ylabel': 'Index' }

        return analysis



from __future__ import division, print_function, absolute_import, unicode_literals
#*****************************************************************
#    pyGSTi 0.9:  Copyright 2015 Sandia Corporation
#    This Software is released under the GPL license detailed
#    in the file "license.txt" in the top-level pyGSTi directory
#*****************************************************************
""" Utility functions related to Gram matrix construction."""

from .. import construction as _construction
from .core import gram_rank_and_evals as _gramRankAndEvals


########################################################
## Gram matrix stuff
########################################################

def get_max_gram_basis(gateLabels, dataset, maxLength=0):
    """
    Compute a maximal set of gate strings that can be used as a basis for a Gram
      matrix.  That is, a maximal set of strings {S_i} such that the gate
      strings { S_i S_j } are all present in dataset.  If maxLength > 0, then
      restrict len(S_i) <= maxLength.

    Parameters
    ----------
    gateLabels : list or tuple
      the gate labels to use in Gram matrix basis strings

    dataset : DataSet
      the dataset to use when constructing the Gram matrix

    maxLength : int, optional
      the maximum string length considered for Gram matrix basis
      elements.  Defaults to 0 (no limit).

    Returns
    -------
    list of tuples
      where each tuple contains gate labels and specifies a single gate string.
    """

    datasetStrings = list(dataset.keys())
    minLength = min( [len(s) for s in datasetStrings] )
    if maxLength <= 0:
        maxLength = max( [len(s) for s in datasetStrings] )
    possibleStrings = _construction.gen_all_gatestrings(gateLabels, (minLength+1)//2, maxLength//2)

    def have_all_data(strings,datasetStrs):
        for a in strings:
            for b in strings:
                if tuple(list(a) + list(b)) not in datasetStrs:
                    return False
        return True

    max_string_set = [ ]
    for p in possibleStrings:
        if have_all_data(max_string_set + [p], datasetStrings):
            max_string_set.append(p)

    return max_string_set


def max_gram_rank_and_evals(dataset, maxBasisStringLength=10,
                            targetGateset=None, spamDict=None,
                            fixedLists=None):
    """
    Compute the rank and singular values of a maximal Gram matrix,that is, the
    Gram matrix using a basis computed by:
    get_max_gram_basis(dataset.get_gate_labels(), dataset, maxBasisStringLength).

    Parameters
    ----------
    dataset : DataSet
      the dataset to use when constructing the Gram matrix

    maxBasisStringLength : int, optional
      the maximum string length considered for Gram matrix basis
      elements.  Defaults to 10.

    targetGateset : GateSet, optional
      A gateset used to specify the SPAM labels used to connect
      the dataset values to rhoVec and EVec indices.

    spamDict : dictionary, optional
      Dictionary mapping (rhoVec_index,EVec_index) integer tuples to string spam labels.
      Defaults to the spam dictionary of targetGateset
      e.g. spamDict[(0,0)] == "plus"

    fixedLists : (prepStrs, effectStrs), optional
      2-tuple of gate string lists, specifying the preparation and
      measurement fiducials to use when constructing the Gram matrix,
      and thereby bypassing the search for such lists.


    Returns
    -------
    rank : integer
    singularvalues : numpy array
    targetsingularvalues : numpy array
    """
    if fixedLists is not None:
        maxRhoStrs, maxEStrs = fixedLists
    else:
        maxStringSet = get_max_gram_basis(dataset.get_gate_labels(), dataset, maxBasisStringLength)
        maxRhoStrs = maxEStrs = maxStringSet

    if spamDict is None:
        if targetGateset is not None:
            spamDict = targetGateset.get_reverse_spam_defs()
            rhoLabels = list(targetGateset.preps.keys())
            eLabels = list(targetGateset.effects.keys()) # 'remainder' should *not* be an effectLabel here
        else:
            firstSpamLabel = dataset.get_spam_labels()[0]
            spamDict = {('dummy_rho','dummy_E'): firstSpamLabel}
            rhoLabels = ['dummy_rho']; eLabels = ['dummy_E']
            # Note: it doesn't actually matter what strings we use here

    specs = _construction.build_spam_specs(prepStrs=maxRhoStrs, effectStrs=maxEStrs,
                                           prep_labels=rhoLabels, effect_labels=eLabels)
    return _gramRankAndEvals(dataset, specs, targetGateset, spamDict)



from __future__ import division, print_function, absolute_import, unicode_literals
#*****************************************************************
#    pyGSTi 0.9:  Copyright 2015 Sandia Corporation
#    This Software is released under the GPL license detailed
#    in the file "license.txt" in the top-level pyGSTi directory
#*****************************************************************
""" Custom implementation of the Levenberg-Marquardt Algorithm """

import time as _time
import numpy as _np
#from scipy.optimize import OptimizeResult as _optResult 

from ..tools import mpitools as _mpit

#constants
MACH_PRECISION = 1e-12


def custom_leastsq(obj_fn, jac_fn, x0, f_norm2_tol=1e-6, jac_norm_tol=1e-6,
                   rel_ftol=1e-6, rel_xtol=1e-6, max_iter=100, comm=None,
                   verbosity=0, profiler=None):
    msg = ""
    converged = False
    x = x0
    f = obj_fn(x)
    norm_f = _np.dot(f,f) # _np.linalg.norm(f)**2
    half_max_nu = 2**62 #what should this be??
    tau = 1e-3
    nu = 2
    mu = 0 #initialized on 1st iter
    my_cols_slice = None

    if comm is not None and comm.Get_rank() != 0:
        verbosity = 0 #Only print to stdout from root process

    if not _np.isfinite(norm_f):
        msg = "Infinite norm of objective function at initial point!"


    for k in range(max_iter): #outer loop
        # assume x, f, fnorm hold valid values

        if len(msg) > 0: 
            break #exit outer loop if an exit-message has been set

        if norm_f < f_norm2_tol:
            msg = "Sum of squares is at most %g" % f_norm2_tol
            converged = True; break

        if verbosity > 0:
            print("--- Outer Iter %d: norm_f = %g, mu=%g" % (k,norm_f,mu))
            
        if profiler: profiler.mem_check("custom_leastsq: begin outer iter *before de-alloc*")
        Jac = None; JTJ = None; JTf = None

        if profiler: profiler.mem_check("custom_leastsq: begin outer iter")
        Jac = jac_fn(x)
        if profiler: profiler.mem_check("custom_leastsq: after jacobian:" 
                                        + "shape=%s, GB=%.2f" % (str(Jac.shape),
                                                        Jac.nbytes/(1024.0**3)) )

        tm = _time.time()
        if my_cols_slice is None:
            my_cols_slice = _mpit.distribute_for_dot(Jac.shape[0], comm)
        JTJ = _mpit.mpidot(Jac.T,Jac,my_cols_slice,comm)   #_np.dot(Jac.T,Jac)
        JTf = _np.dot(Jac.T,f)
        if profiler: profiler.add_time("custom_leastsq: dotprods",tm)

        idiag = _np.diag_indices_from(JTJ)
        norm_JTf = _np.linalg.norm(JTf,ord=_np.inf)
        norm_x = _np.dot(x,x) # _np.linalg.norm(x)**2
        undampled_JTJ_diag = JTJ.diagonal().copy()

        if norm_JTf < jac_norm_tol:
            msg = "norm(jacobian) is at most %g" % jac_norm_tol
            converged = True; break

        if k == 0:
            #mu = tau # initial damping element
            mu = tau * _np.max(undampled_JTJ_diag) # initial damping element

        #determing increment using adaptive damping
        while True:  #inner loop

            if profiler: profiler.mem_check("custom_leastsq: begin inner iter")
            JTJ[idiag] += mu # augment normal equations
            #JTJ[idiag] *= (1.0 + mu) # augment normal equations

            try:
                if profiler: profiler.mem_check("custom_leastsq: before linsolve")
                tm = _time.time()
                success = True
                dx = _np.linalg.solve(JTJ, -JTf) 
                if profiler: profiler.add_time("custom_leastsq: linsolve",tm)
            except _np.linalg.LinAlgError:
                success = False
            
            if profiler: profiler.mem_check("custom_leastsq: after linsolve")
            if success: #linear solve succeeded
                new_x = x + dx
                norm_dx = _np.dot(dx,dx) # _np.linalg.norm(dx)**2

                if verbosity > 1:
                    print("  - Inner Loop: mu=%g, norm_dx=%g" % (mu,norm_dx))

                if norm_dx < (rel_xtol**2)*norm_x:
                    msg = "Relative change in |x| is at most %g" % rel_xtol
                    converged = True; break

                if norm_dx > (norm_x+rel_xtol)/(MACH_PRECISION**2):
                    msg = "(near-)singular linear system"; break
                
                new_f = obj_fn(new_x)
                if profiler: profiler.mem_check("custom_leastsq: after obj_fn")
                norm_new_f = _np.dot(new_f,new_f) # _np.linalg.norm(new_f)**2
                if not _np.isfinite(norm_new_f): # avoid infinite loop...
                    msg = "Infinite norm of objective function!"; break

                dL = _np.dot(dx, mu*dx - JTf) # expected decrease in ||F||^2 from linear model
                dF = norm_f - norm_new_f      # actual decrease in ||F||^2

                if verbosity > 1:
                    print("      (cont): norm_new_f=%g, dL=%g, dF=%g, reldL=%g, reldF=%g" % 
                          (norm_new_f,dL,dF,dL/norm_f,dF/norm_f))

                if dL/norm_f < rel_ftol and dF/norm_f < rel_ftol and dF/dL < 2.0:
                    msg = "Both actual and predicted relative reductions in the" + \
                        " sum of squares are at most %g" % rel_ftol
                    converged = True; break

                if profiler: profiler.mem_check("custom_leastsq: before success")

                if dL > 0 and dF > 0:
                    # reduction in error: increment accepted!
                    t = 1.0 - (2*dF/dL-1.0)**3 # dF/dL == gain ratio
                    mu *= max(t,1.0/3.0)
                    nu = 2
                    x,f, norm_f = new_x, new_f, norm_new_f

                    if verbosity > 1:
                        print("      Accepted! gain ratio=%g  mu * %g => %g"
                              % (dF/dL,max(t,1.0/3.0),mu))

                    ##Check to see if we *would* switch to Q-N method in a hybrid algorithm
                    #new_Jac = jac_fn(new_x)
                    #new_JTf = _np.dot(new_Jac.T,new_f)
                    #print(" CHECK: %g < %g ?" % (_np.linalg.norm(new_JTf,
                    #    ord=_np.inf),0.02 * _np.linalg.norm(new_f)))

                    break # exit inner loop normally
            #else:
            #    print("LinSolve Failure!!")

            # if this point is reached, either the linear solve failed
            # or the error did not reduce.  In either case, reject increment.
                
            #Increase damping (mu), then increase damping factor to 
            # accelerate further damping increases.
            mu *= nu
            if nu > half_max_nu : #watch for nu getting too large (&overflow)
                msg = "Stopping after nu overflow!"; break
            nu = 2*nu
            if verbosity > 1:
                print("      Rejected!  mu => mu*nu = %g, nu => 2*nu = %g"
                      % (mu, nu))
            
            JTJ[idiag] = undampled_JTJ_diag #restore diagonal
        #end of inner loop
    #end of outer loop
    else:
        #if no break stmt hit, then we've exceeded maxIter
        msg = "Maximum iterations (%d) exceeded" % max_iter

    #JTJ[idiag] = undampled_JTJ_diag #restore diagonal
    return x, converged, msg
    #solution = _optResult()
    #solution.x = x; solution.fun = f
    #solution.success = converged
    #solution.message = msg
    #return solution



#Wikipedia-version of LM algorithm, testing mu and mu/nu damping params and taking 
# mu/nu => new_mu if acceptable...  This didn't seem to perform well, but maybe just
# needs some tweaking, so leaving it commented here for reference
#def custom_leastsq_wikip(obj_fn, jac_fn, x0, f_norm_tol=1e-6, jac_norm_tol=1e-6,
#                   rel_tol=1e-6, max_iter=100, comm=None, verbosity=0, profiler=None):
#    msg = ""
#    converged = False
#    x = x0
#    f = obj_fn(x)
#    norm_f = _np.linalg.norm(f)
#    tau = 1e-3 #initial mu
#    nu = 1.3
#    my_cols_slice = None
#    
#
#    if not _np.isfinite(norm_f):
#        msg = "Infinite norm of objective function at initial point!"
#
#    for k in range(max_iter): #outer loop
#        # assume x, f, fnorm hold valid values
#
#        if len(msg) > 0: 
#            break #exit outer loop if an exit-message has been set
#
#        if norm_f < f_norm_tol:
#            msg = "norm(objectivefn) is small"
#            converged = True; break
#
#        if verbosity > 0:
#            print("--- Outer Iter %d: norm_f = %g" % (k,norm_f))
#            
#        if profiler: profiler.mem_check("custom_leastsq: begin outer iter *before de-alloc*")
#        Jac = None; JTJ = None; JTf = None
#
#        if profiler: profiler.mem_check("custom_leastsq: begin outer iter")
#        Jac = jac_fn(x)
#        if profiler: profiler.mem_check("custom_leastsq: after jacobian:" 
#                                        + "shape=%s, GB=%.2f" % (str(Jac.shape),
#                                                        Jac.nbytes/(1024.0**3)) )
#
#        tm = _time.time()
#        if my_cols_slice is None:
#            my_cols_slice = _mpit.distribute_for_dot(Jac.shape[0], comm)
#        JTJ = _mpit.mpidot(Jac.T,Jac,my_cols_slice,comm)   #_np.dot(Jac.T,Jac)
#        JTf = _np.dot(Jac.T,f)
#        if profiler: profiler.add_time("custom_leastsq: dotprods",tm)
#
#        idiag = _np.diag_indices_from(JTJ)
#        norm_JTf = _np.linalg.norm(JTf) #, ord='inf')
#        norm_x = _np.linalg.norm(x)
#        undampled_JTJ_diag = JTJ.diagonal().copy()
#
#        if norm_JTf < jac_norm_tol:
#            msg = "norm(jacobian) is small"
#            converged = True; break
#
#        if k == 0:
#            mu = tau #* _np.max(undampled_JTJ_diag) # initial damping element
#        #mu = tau #* _np.max(undampled_JTJ_diag) # initial damping element
#
#        #determing increment using adaptive damping
#        while True:  #inner loop
#
#            ### Evaluate with mu' = mu / nu
#            mu = mu / nu
#            if profiler: profiler.mem_check("custom_leastsq: begin inner iter")
#            JTJ[idiag] *= (1.0 + mu) # augment normal equations
#            #JTJ[idiag] += mu # augment normal equations
#
#            try:
#                if profiler: profiler.mem_check("custom_leastsq: before linsolve")
#                tm = _time.time()
#                success = True
#                dx = _np.linalg.solve(JTJ, -JTf) 
#                if profiler: profiler.add_time("custom_leastsq: linsolve",tm)
#            except _np.linalg.LinAlgError:
#                success = False
#            
#            if profiler: profiler.mem_check("custom_leastsq: after linsolve")
#            if success: #linear solve succeeded
#                new_x = x + dx
#                norm_dx = _np.linalg.norm(dx)
#
#                #if verbosity > 1:
#                #    print("--- Inner Loop: mu=%g, norm_dx=%g" % (mu,norm_dx))
#
#                if norm_dx < rel_tol*norm_x: #use squared qtys instead (speed)?
#                    msg = "relative change in x is small"
#                    converged = True; break
#
#                if norm_dx > (norm_x+rel_tol)/MACH_PRECISION:
#                    msg = "(near-)singular linear system"; break
#                
#                new_f = obj_fn(new_x)
#                if profiler: profiler.mem_check("custom_leastsq: after obj_fn")
#                norm_new_f = _np.linalg.norm(new_f)
#                if not _np.isfinite(norm_new_f): # avoid infinite loop...
#                    msg = "Infinite norm of objective function!"; break
#
#                dF = norm_f - norm_new_f
#                if dF > 0: #accept step
#                    #print("      Accepted!")
#                    x,f, norm_f = new_x, new_f, norm_new_f
#                    nu = 1.3
#                    break # exit inner loop normally
#                else:
#                    mu *= nu #increase mu
#            else:
#                #Linear solve failed:
#                mu *= nu #increase mu
#                nu = 2*nu
#            
#            JTJ[idiag] = undampled_JTJ_diag #restore diagonal for next inner loop iter
#        #end of inner loop
#    #end of outer loop
#    else:
#        #if no break stmt hit, then we've exceeded maxIter
#        msg = "Maximum iterations (%d) exceeded" % max_iter
#
#    return x, converged, msg



from __future__ import division, print_function, absolute_import, unicode_literals
#*****************************************************************
#    pyGSTi 0.9:  Copyright 2015 Sandia Corporation
#    This Software is released under the GPL license detailed
#    in the file "license.txt" in the top-level pyGSTi directory
#*****************************************************************
""" Functions for creating datasets """

import numpy as _np
import numpy.random as _rndm
import warnings as _warnings

from ..objects import gatestring as _gs
from ..objects import dataset as _ds
from . import gatestringconstruction as _gstrc

def generate_fake_data(gatesetOrDataset, gatestring_list, nSamples,
                       sampleError="none", seed=None, randState=None,
                       aliasDict=None, collisionAction="aggregate"):
    """Creates a DataSet using the probabilities obtained from a gateset.

    Parameters
    ----------
    gatesetOrDataset : GateSet or DataSet object
        If a GateSet, the gate set whose probabilities generate the data.
        If a DataSet, the data set whose frequencies generate the data.

    gatestring_list : list of (tuples or GateStrings) or None
        Each tuple or GateString contains gate labels and
        specifies a gate sequence whose counts are included
        in the returned DataSet. e.g. ``[ (), ('Gx',), ('Gx','Gy') ]``

    nSamples : int or list of ints or None
        The simulated number of samples for each gate string.  This only has
        effect when  ``sampleError == "binomial"`` or ``"multinomial"``.  If an
        integer, all gate strings have this number of total samples. If a list,
        integer elements specify the number of samples for the corresponding
        gate string.  If ``None``, then `gatesetOrDataset` must be a
        :class:`~pygsti.objects.DataSet`, and total counts are taken from it
        (on a per-gatestring basis).

    sampleError : string, optional
        What type of sample error is included in the counts.  Can be:

        - "none"  - no sample error: counts are floating point numbers such
          that the exact probabilty can be found by the ratio of count / total.
        - "round" - same as "none", except counts are rounded to the nearest
          integer.
        - "binomial" - the number of counts is taken from a binomial
          distribution.  Distribution has parameters p = probability of the
          gate string and n = number of samples.  This can only be used when
          there are exactly two SPAM labels in gatesetOrDataset.
        - "multinomial" - counts are taken from a multinomial distribution.
          Distribution has parameters p_k = probability of the gate string
          using the k-th SPAM label and n = number of samples.

    seed : int, optional
        If not ``None``, a seed for numpy's random number generator, which
        is used to sample from the binomial or multinomial distribution.

    randState : numpy.random.RandomState
        A RandomState object to generate samples from. Can be useful to set
        instead of `seed` if you want reproducible distribution samples across
        multiple random function calls but you don't want to bother with
        manually incrementing seeds between those calls.

    aliasDict : dict, optional
        A dictionary mapping single gate labels into tuples of one or more 
        other gate labels which translate the given gate strings before values
        are computed using `gatesetOrDataset`.  The resulting Dataset, however,
        contains the *un-translated* gate strings as keys.

    collisionAction : {"aggregate", "keepseparate"}
        Determines how duplicate gate sequences are handled by the resulting
        `DataSet`.  Please see the constructor documentation for `DataSet`.

    Returns
    -------
    DataSet
       A static data set filled with counts for the specified gate strings.

    """
    if isinstance(gatesetOrDataset, _ds.DataSet):
        dsGen = gatesetOrDataset #dataset
        gsGen = None
        dataset = _ds.DataSet( spamLabels=dsGen.get_spam_labels(),
                               collisionAction=collisionAction)
    else:
        gsGen = gatesetOrDataset #dataset
        dsGen = None
        dataset = _ds.DataSet( spamLabels=gsGen.get_spam_labels(),
                               collisionAction=collisionAction )

    if sampleError in ("binomial","multinomial"):
        if randState is None:
            rndm = _rndm.RandomState(seed) # ok if seed is None
        else:
            rndm = randState

    if aliasDict is not None:
        translated_list = _gstrc.translate_gatestring_list(
                                      gatestring_list, aliasDict)
    else: translated_list = gatestring_list

    for k,(s,trans_s) in enumerate(zip(gatestring_list,translated_list)):

        if gsGen:
            ps = gsGen.probs(trans_s) 
              # a dictionary of probabilities; keys = spam labels

            if sampleError in ("binomial","multinomial"):
                #Adjust to probabilities if needed (and warn if not close to in-bounds)
                TOL = 1e-10
                for sl in ps: 
                    if ps[sl] < 0:
                        if ps[sl] < -TOL: _warnings.warn("Clipping probs < 0 to 0")
                        ps[sl] = 0.0
                    elif ps[sl] > 1: 
                        if ps[sl] > (1+TOL): _warnings.warn("Clipping probs > 1 to 1")
                        ps[sl] = 1.0
                
                psum = sum(ps.values())
                if psum > 1:
                    if psum > 1+TOL: _warnings.warn("Adjusting sum(probs) > 1 to 1")
                    extra_p = (psum-1.0) * (1.000000001) # to sum < 1+eps (numerical prec insurance)
                    for sl in ps:
                        if extra_p > 0:
                            x = min(ps[sl],extra_p)
                            ps[sl] -= x; extra_p -= x
                        else: break
                    
                assert(-TOL <= sum(ps.values()) <= 1.+TOL)
        else:
            ps = { sl: dsGen[trans_s].fraction(sl) 
                   for sl in dsGen.get_spam_labels() }

        if nSamples is None and dsGen is not None:
            N = dsGen[trans_s].total() #use the number of samples from the generating dataset
        else:
            try:
                N = nSamples[k] #try to treat nSamples as a list
            except:
                N = nSamples #if not indexable, nSamples should be a single number

        #Weight the number of samples according to a WeightedGateString
        if isinstance(s, _gs.WeightedGateString):
            nWeightedSamples = int(round(s.weight * N))
        else:
            nWeightedSamples = N

        counts = { }
        if sampleError == "binomial":
            assert(len(list(ps.keys())) == 2)
            spamLabel1, spamLabel2 = sorted(list(ps.keys())); p1 = ps[spamLabel1]
            counts[spamLabel1] = rndm.binomial(nWeightedSamples, p1) #numpy.clip(p1,0,1) )
            counts[spamLabel2] = nWeightedSamples - counts[spamLabel1]
        elif sampleError == "multinomial":
            #nOutcomes = len(list(ps.keys()))
            spamLabels = list(ps.keys())
            countsArray = rndm.multinomial(nWeightedSamples,
                    [ps[sl] for sl in spamLabels], size=1)
            for i,spamLabel in enumerate(spamLabels):
                counts[spamLabel] = countsArray[0,i]
        else:
            for (spamLabel,p) in ps.items():
                pc = _np.clip(p,0,1)
                if sampleError == "none":
                    counts[spamLabel] = float(nWeightedSamples * pc)
                elif sampleError == "round":
                    counts[spamLabel] = int(round(nWeightedSamples*pc))
                else: raise ValueError("Invalid sample error parameter: '%s'  Valid options are 'none', 'round', 'binomial', or 'multinomial'" % sampleError)

        dataset.add_count_dict(s, counts)
    dataset.done_adding_data()
    return dataset
    
def merge_outcomes(dataset,label_merge_dict):
    """Creates a DataSet which merges certain outcomes in input DataSet;
    used, for example, to aggregate a 2-qubit 4-outcome DataSet into a 1-qubit 2-outcome
    DataSet.
    
    Parameters
    ----------
    dataset : DataSet object
        The input DataSet whose results will be compiled according to the rules 
        set forth in label_merge_dict

    label_merge_dict : dictionary
        The dictionary whose keys define the new DataSet outcomes, and whose items 
        are lists of input DataSet outcomes that are to be summed together.  For example,
        if a two-qubit DataSet has outcome labels "upup", "updn", "dnup", and "dndn", and
        we want to ''trace out'' the second qubit, we could use label_merge_dict =
        {'plus':['upup','updn'],'minus':['dnup','dndn']}.
    
    Returns
    -------
    merged_dataset : DataSet object
        The DataSet with outcomes merged according to the rules given in label_merge_dict.
    """

    new_effects = label_merge_dict.keys()
    merged_dataset = _ds.DataSet(spamLabels=new_effects)
    if sorted([effect for sublist in label_merge_dict.values() for effect in sublist]) != sorted(dataset.get_spam_labels()):
        print('Warning: There is a mismatch between original effects in label_merge_dict and original effects in original dataset.')
    for key in dataset.keys():
        dataline = dataset[key]
        count_dict = {}
        for new_effect in new_effects:
            count_dict[new_effect] = 0
            for old_effect in label_merge_dict[new_effect]:
                count_dict[new_effect] += dataline[old_effect]
        merged_dataset.add_count_dict(key,count_dict)
    merged_dataset.done_adding_data()
    return merged_dataset



from __future__ import division, print_function, absolute_import, unicode_literals
#*****************************************************************
#    pyGSTi 0.9:  Copyright 2015 Sandia Corporation
#    This Software is released under the GPL license detailed
#    in the file "license.txt" in the top-level pyGSTi directory
#*****************************************************************
"""
Functions which compute named quantities for GateSets and Datasets.

Named quantities as well as their confidence-region error bars are
 computed by the functions in this module. These quantities are
 used primarily in reports, so we refer to these quantities as
 "reportables".
"""
import numpy as _np
from collections import OrderedDict as _OrderedDict

from .. import tools as _tools
from .. import algorithms as _alg

FINITE_DIFF_EPS = 1e-7

class ReportableQty(object):
    """
    Encapsulates a computed quantity and possibly its error bars,
    primarily for use in reports.
    """

    def __init__(self, value, errbar=None):
        """
        Initialize a new ReportableQty object, which
        is essentially a container for a value and error bars.

        Parameters
        ----------
        value : anything
           The value to store

        errbar : anything
           The error bar(s) to store
        """
        self.value = value
        self.errbar = errbar

    def get_value(self):
        """
        Returns the quantity's value
        """
        return self.value

    def get_err_bar(self):
        """
        Returns the quantity's error bar(s)
        """
        return self.errbar

    def get_value_and_err_bar(self):
        """
        Returns the quantity's value and error bar(s)
        """
        return self.value, self.errbar

    def __str__(self):
        if self.errbar is not None:
            return str(self.value) + " +/- " + str(self.errbar)
        else: return str(self.value)


def _projectToValidProb(p, tol=1e-9):
    if p < tol: return tol
    if p > 1-tol: return 1-tol
    return p


def _getGateQuantity(fnOfGate, gateset, gateLabel, eps, confidenceRegionInfo, verbosity=0):
    """ For constructing a ReportableQty from a function of a gate. """

    if confidenceRegionInfo is None: # No Error bars
        return ReportableQty(fnOfGate(gateset.gates[gateLabel]))

    # make sure the gateset we're given is the one used to generate the confidence region
    if(gateset.frobeniusdist(confidenceRegionInfo.get_gateset()) > 1e-6):
        raise ValueError("Prep quantity confidence region is being requested for " +
                         "a different gateset than the given confidenceRegionInfo")

    df, f0 = confidenceRegionInfo.get_gate_fn_confidence_interval(fnOfGate, gateLabel,
                                                              eps, returnFnVal=True,
                                                              verbosity=verbosity)
    return ReportableQty(f0,df)

def _getPrepQuantity(fnOfPrep, gateset, prepLabel, eps, confidenceRegionInfo, verbosity=0):
    """ For constructing a ReportableQty from a function of a state preparation. """

    if confidenceRegionInfo is None: # No Error bars
        return ReportableQty(fnOfPrep(gateset.preps[prepLabel]))

    # make sure the gateset we're given is the one used to generate the confidence region
    if(gateset.frobeniusdist(confidenceRegionInfo.get_gateset()) > 1e-6):
        raise ValueError("Gate quantity confidence region is being requested for " +
                         "a different gateset than the given confidenceRegionInfo")

    df, f0 = confidenceRegionInfo.get_prep_fn_confidence_interval(fnOfPrep, prepLabel,
                                                              eps, returnFnVal=True,
                                                              verbosity=verbosity)
    return ReportableQty(f0,df)


def _getEffectQuantity(fnOfEffect, gateset, effectLabel, eps, confidenceRegionInfo, verbosity=0):
    """ For constructing a ReportableQty from a function of a POVM effect. """

    if confidenceRegionInfo is None: # No Error bars
        return ReportableQty(fnOfEffect(gateset.effects[effectLabel]))

    # make sure the gateset we're given is the one used to generate the confidence region
    if(gateset.frobeniusdist(confidenceRegionInfo.get_gateset()) > 1e-6):
        raise ValueError("Effect quantity confidence region is being requested for " +
                         "a different gateset than the given confidenceRegionInfo")

    df, f0 = confidenceRegionInfo.get_effect_fn_confidence_interval(
        fnOfEffect, effectLabel, eps, returnFnVal=True, verbosity=verbosity)
    return ReportableQty(f0,df)


def _getGateSetQuantity(fnOfGateSet, gateset, eps, confidenceRegionInfo, verbosity=0):
    """ For constructing a ReportableQty from a function of a gate. """

    if confidenceRegionInfo is None: # No Error bars
        return ReportableQty(fnOfGateSet(gateset))

    # make sure the gateset we're given is the one used to generate the confidence region
    if(gateset.frobeniusdist(confidenceRegionInfo.get_gateset()) > 1e-6):
        raise ValueError("GateSet quantity confidence region is being requested for " +
                         "a different gateset than the given confidenceRegionInfo")

    df, f0 = confidenceRegionInfo.get_gateset_fn_confidence_interval(
        fnOfGateSet, eps, returnFnVal=True, verbosity=verbosity)

    return ReportableQty(f0,df)


def _getSpamQuantity(fnOfSpamVecs, gateset, eps, confidenceRegionInfo, verbosity=0):
    """ For constructing a ReportableQty from a function of a spam vectors."""

    if confidenceRegionInfo is None: # No Error bars
        return ReportableQty(fnOfSpamVecs(gateset.get_preps(), gateset.get_effects()))

    # make sure the gateset we're given is the one used to generate the confidence region
    if(gateset.frobeniusdist(confidenceRegionInfo.get_gateset()) > 1e-6):
        raise ValueError("Spam quantity confidence region is being requested for " +
                         "a different gateset than the given confidenceRegionInfo")

    df, f0 = confidenceRegionInfo.get_spam_fn_confidence_interval(fnOfSpamVecs,
                                                              eps, returnFnVal=True,
                                                              verbosity=verbosity)
    return ReportableQty(f0,df)



def compute_dataset_qty(qtyname, dataset, gatestrings=None):
    """
    Compute the named "Dataset" quantity.

    Parameters
    ----------
    qtyname : string
        Name of the quantity to compute.

    dataset : DataSet
        Data used to compute the quantity.

    gatestrings : list of tuples or GateString objects, optional
        A list of gatestrings used in the computation of certain quantities.
        If None, all the gatestrings in the dataset are used.

    Returns
    -------
    ReportableQty
        The quantity requested, or None if quantity could not be computed.
    """
    ret = compute_dataset_qtys( [qtyname], dataset, gatestrings )
    if qtyname is None: return ret
    elif qtyname in ret: return ret[qtyname]
    else: return None

def compute_dataset_qtys(qtynames, dataset, gatestrings=None):
    """
    Compute the named "Dataset" quantities.

    Parameters
    ----------
    qtynames : list of strings
        Names of the quantities to compute.

    dataset : DataSet
        Data used to compute the quantity.

    gatestrings : list of tuples or GateString objects, optional
        A list of gatestrings used in the computation of certain quantities.
        If None, all the gatestrings in the dataset are used.

    Returns
    -------
    dict
        Dictionary whose keys are the requested quantity names and values are
        ReportableQty objects.
    """

    ret = _OrderedDict()
    possible_qtys = [ ]

    #Quantities computed per gatestring
    per_gatestring_qtys = _OrderedDict( [('gate string', []), ('gate string index', []), ('gate string length', []), ('count total', [])] )
    spamLabels = dataset.get_spam_labels()
    for spl in spamLabels:
        per_gatestring_qtys['Exp prob(%s)' % spl] = []
        per_gatestring_qtys['Exp count(%s)' % spl] = []

    if any( [qtyname in per_gatestring_qtys for qtyname in qtynames ] ):
        if gatestrings is None: gatestrings = list(dataset.keys())
        for (i,gs) in enumerate(gatestrings):
            if gs in dataset: # skip gate strings given that are not in dataset
                dsRow = dataset[gs]
            else:
                #print "Warning: skipping gate string %s" % str(gs)
                continue

            N = dsRow.total()
            per_gatestring_qtys['gate string'].append(  ''.join(gs)  )
            per_gatestring_qtys['gate string index'].append( i )
            per_gatestring_qtys['gate string length'].append(  len(gs)  )
            per_gatestring_qtys['count total'].append(  N  )

            for spamLabel in spamLabels:
                pExp = _projectToValidProb( dsRow[spamLabel] / N, tol=1e-10 )
                per_gatestring_qtys['Exp prob(%s)' % spamLabel].append( pExp )
                per_gatestring_qtys['Exp count(%s)' % spamLabel].append( dsRow[spamLabel] )

        for qtyname in qtynames:
            if qtyname in per_gatestring_qtys:
                ret[qtyname] = ReportableQty(per_gatestring_qtys[qtyname])


    #Quantities computed per dataset
    qty = "max logl"; possible_qtys.append(qty)
    if qty in qtynames:
        ret[qty] = ReportableQty( _tools.logl_max(dataset))

    qty = "number of gate strings"; possible_qtys.append(qty)
    if qty in qtynames:
        ret[qty] = ReportableQty( len(dataset) )

    if qtynames[0] is None:
        return possible_qtys + list(per_gatestring_qtys.keys())
    return ret


def compute_gateset_qty(qtyname, gateset, confidenceRegionInfo=None):
    """
    Compute the named "GateSet" quantity.

    Parameters
    ----------
    qtyname : string
        Name of the quantity to compute.

    gateset : GateSet
        Gate set used to compute the quantity.

    confidenceRegionInfo : ConfidenceRegion, optional
        If not None, specifies a confidence-region used to compute the error bars
        contained in the returned quantity.  If None, then no error bars are
        computed.

    Returns
    -------
    ReportableQty
        The quantity requested, or None if quantity could not be computed.
    """
    ret = compute_gateset_qtys( [qtyname], gateset, confidenceRegionInfo)
    if qtyname is None: return ret
    elif qtyname in ret: return ret[qtyname]
    else: return None

def compute_gateset_qtys(qtynames, gateset, confidenceRegionInfo=None):
    """
    Compute the named "GateSet" quantities.

    Parameters
    ----------
    qtynames : list of strings
        Names of the quantities to compute.

    gateset : GateSet
        Gate set used to compute the quantities.

    confidenceRegionInfo : ConfidenceRegion, optional
        If not None, specifies a confidence-region used to compute the error bars
        contained in the returned quantities.  If None, then no error bars are
        computed.

    Returns
    -------
    dict
        Dictionary whose keys are the requested quantity names and values are
        ReportableQty objects.
    """
    ret = _OrderedDict()
    possible_qtys = [ ]
    eps = FINITE_DIFF_EPS
    mxBasis = gateset.get_basis_name()

    def choi_matrix(gate):
        return _tools.jamiolkowski_iso(gate, mxBasis, mxBasis)

    def choi_evals(gate):
        choi = _tools.jamiolkowski_iso(gate, mxBasis, mxBasis)
        choi_eigvals = _np.linalg.eigvals(choi)
        return _np.array(sorted(choi_eigvals))

    def choi_trace(gate):
        choi = _tools.jamiolkowski_iso(gate, mxBasis, mxBasis)
        return _np.trace(choi)

    def decomp_angle(gate):
        decomp = _tools.decompose_gate_matrix(gate)
        return decomp.get('pi rotations',0)

    def decomp_decay_diag(gate):
        decomp = _tools.decompose_gate_matrix(gate)
        return decomp.get('decay of diagonal rotation terms',0)

    def decomp_decay_offdiag(gate):
        decomp = _tools.decompose_gate_matrix(gate)
        return decomp.get('decay of off diagonal rotation terms',0)

    def decomp_cu_angle(gate):
        closestUGateMx = _alg.find_closest_unitary_gatemx(gate)
        decomp = _tools.decompose_gate_matrix(closestUGateMx)
        return decomp.get('pi rotations',0)

    def decomp_cu_decay_diag(gate):
        closestUGateMx = _alg.find_closest_unitary_gatemx(gate)
        decomp = _tools.decompose_gate_matrix(closestUGateMx)
        return decomp.get('decay of diagonal rotation terms',0)

    def decomp_cu_decay_offdiag(gate):
        closestUGateMx = _alg.find_closest_unitary_gatemx(gate)
        decomp = _tools.decompose_gate_matrix(closestUGateMx)
        return decomp.get('decay of off diagonal rotation terms',0)

    def upper_bound_fidelity(gate):
        return _tools.fidelity_upper_bound(gate)[0]

    def closest_ujmx(gate):
        closestUGateMx = _alg.find_closest_unitary_gatemx(gate)
        return _tools.jamiolkowski_iso(closestUGateMx, mxBasis, mxBasis)

    def maximum_fidelity(gate):
        closestUGateMx = _alg.find_closest_unitary_gatemx(gate)
        closestUJMx = _tools.jamiolkowski_iso(closestUGateMx, mxBasis, mxBasis)
        choi = _tools.jamiolkowski_iso(gate, mxBasis, mxBasis)
        return _tools.fidelity(closestUJMx, choi)

    def maximum_trace_dist(gate):
        closestUGateMx = _alg.find_closest_unitary_gatemx(gate)
        #closestUJMx = _tools.jamiolkowski_iso(closestUGateMx, mxBasis, mxBasis)
        _tools.jamiolkowski_iso(closestUGateMx, mxBasis, mxBasis)
        return _tools.jtracedist(gate, closestUGateMx)

    def spam_dotprods(rhoVecs, EVecs):
        ret = _np.empty( (len(rhoVecs), len(EVecs)), 'd')
        for i,rhoVec in enumerate(rhoVecs):
            for j,EVec in enumerate(EVecs):
                ret[i,j] = _np.dot(_np.transpose(EVec), rhoVec)
        return ret

    def angles_btwn_rotn_axes(gateset):
        gateLabels = list(gateset.gates.keys())
        angles_btwn_rotn_axes = _np.zeros( (len(gateLabels), len(gateLabels)), 'd' )

        for i,gl in enumerate(gateLabels):
            decomp = _tools.decompose_gate_matrix(gateset.gates[gl])
            rotnAngle = decomp.get('pi rotations','X')
            axisOfRotn = decomp.get('axis of rotation',None)

            for j,gl_other in enumerate(gateLabels[i+1:],start=i+1):
                decomp_other = _tools.decompose_gate_matrix(gateset.gates[gl_other])
                rotnAngle_other = decomp_other.get('pi rotations','X')

                if str(rotnAngle) == 'X' or abs(rotnAngle) < 1e-4 or \
                   str(rotnAngle_other) == 'X' or abs(rotnAngle_other) < 1e-4:
                    angles_btwn_rotn_axes[i,j] =  _np.nan
                else:
                    axisOfRotn_other = decomp_other.get('axis of rotation',None)
                    if axisOfRotn is not None and axisOfRotn_other is not None:
                        real_dot =  _np.clip( _np.real(_np.dot(axisOfRotn,axisOfRotn_other)), -1.0, 1.0)
                        angles_btwn_rotn_axes[i,j] = _np.arccos( real_dot ) / _np.pi
                    else:
                        angles_btwn_rotn_axes[i,j] = _np.nan

                angles_btwn_rotn_axes[j,i] = angles_btwn_rotn_axes[i,j]
        return angles_btwn_rotn_axes



    # Spam quantities (computed for all spam vectors at once):
    key = "Spam DotProds"; possible_qtys.append(key)
    if key in qtynames:
        ret[key] = _getSpamQuantity(spam_dotprods, gateset, eps, confidenceRegionInfo)

    key = "Gateset Axis Angles"; possible_qtys.append(key)
    if key in qtynames:
        ret[key] = _getGateSetQuantity(angles_btwn_rotn_axes, gateset, eps, confidenceRegionInfo)

    # Quantities computed per gate
    for (label,gate) in gateset.gates.items():

        #Gate quantities
        suffixes = ('eigenvalues', 'eigenvectors', 'choi eigenvalues', 'choi trace',
                    'choi matrix', 'decomposition')
        gate_qtys = _OrderedDict( [ ("%s %s" % (label,s), None) for s in suffixes ] )
        possible_qtys += list(gate_qtys.keys())

        if any( [qtyname in gate_qtys for qtyname in qtynames] ):
            #gate_evals,gate_evecs = _np.linalg.eig(gate)
            evalsQty = _getGateQuantity(_np.linalg.eigvals, gateset, label, eps, confidenceRegionInfo)
            choiQty = _getGateQuantity(choi_matrix, gateset, label, eps, confidenceRegionInfo)
            choiEvQty = _getGateQuantity(choi_evals, gateset, label, eps, confidenceRegionInfo)
            choiTrQty = _getGateQuantity(choi_trace, gateset, label, eps, confidenceRegionInfo)

            decompDict = _tools.decompose_gate_matrix(gate)
            if decompDict['isValid']:
                angleQty = _getGateQuantity(decomp_angle, gateset, label, eps, confidenceRegionInfo)
                diagQty = _getGateQuantity(decomp_decay_diag, gateset, label, eps, confidenceRegionInfo)
                offdiagQty = _getGateQuantity(decomp_decay_offdiag, gateset, label, eps, confidenceRegionInfo)
                errBarDict = { 'pi rotations': angleQty.get_err_bar(),
                               'decay of diagonal rotation terms': diagQty.get_err_bar(),
                               'decay of off diagonal rotation terms': offdiagQty.get_err_bar() }
                decompQty = ReportableQty(decompDict, errBarDict)
            else:
                decompQty = ReportableQty({})

            gate_qtys[ '%s eigenvalues' % label ]      = evalsQty
            #gate_qtys[ '%s eigenvectors' % label ]     = gate_evecs
            gate_qtys[ '%s choi matrix' % label ]      = choiQty
            gate_qtys[ '%s choi eigenvalues' % label ] = choiEvQty
            gate_qtys[ '%s choi trace' % label ]       = choiTrQty
            gate_qtys[ '%s decomposition' % label]     = decompQty

            for qtyname in qtynames:
                if qtyname in gate_qtys:
                    ret[qtyname] = gate_qtys[qtyname]


        #Closest unitary quantities
        suffixes = ('max fidelity with unitary',
                    'max trace dist with unitary',
                    'upper bound on fidelity with unitary',
                    'closest unitary choi matrix',
                    'closest unitary decomposition')
        closestU_qtys = _OrderedDict( [ ("%s %s" % (label,s), None) for s in suffixes ] )
        possible_qtys += list(closestU_qtys.keys())
        if any( [qtyname in closestU_qtys for qtyname in qtynames] ):
            ubFQty = _getGateQuantity(upper_bound_fidelity, gateset, label, eps, confidenceRegionInfo)
            closeUJMxQty = _getGateQuantity(closest_ujmx, gateset, label, eps, confidenceRegionInfo)
            maxFQty = _getGateQuantity(maximum_fidelity, gateset, label, eps, confidenceRegionInfo)
            maxJTDQty = _getGateQuantity(maximum_trace_dist, gateset, label, eps, confidenceRegionInfo)

            closestUGateMx = _alg.find_closest_unitary_gatemx(gate)
            decompDict = _tools.decompose_gate_matrix(closestUGateMx)
            if decompDict['isValid']:
                angleQty = _getGateQuantity(decomp_cu_angle, gateset, label, eps, confidenceRegionInfo)
                diagQty = _getGateQuantity(decomp_cu_decay_diag, gateset, label, eps, confidenceRegionInfo)
                offdiagQty = _getGateQuantity(decomp_cu_decay_offdiag, gateset, label, eps, confidenceRegionInfo)
                errBarDict = { 'pi rotations': angleQty.get_err_bar(),
                               'decay of diagonal rotation terms': diagQty.get_err_bar(),
                               'decay of off diagonal rotation terms': offdiagQty.get_err_bar() }
                decompQty = ReportableQty(decompDict, errBarDict)
            else:
                decompQty = ReportableQty({})

            closestU_qtys[ '%s max fidelity with unitary' % label ]                  = maxFQty
            closestU_qtys[ '%s max trace dist with unitary' % label ]                = maxJTDQty
            closestU_qtys[ '%s upper bound on fidelity with unitary' % label ]       = ubFQty
            closestU_qtys[ '%s closest unitary choi matrix' % label ]                = closeUJMxQty
            closestU_qtys[ '%s closest unitary decomposition' % label ]              = decompQty

            for qtyname in qtynames:
                if qtyname in closestU_qtys:
                    ret[qtyname] = closestU_qtys[qtyname]

    if qtynames[0] is None:
        return possible_qtys
    return ret


def compute_gateset_dataset_qty(qtyname, gateset, dataset, gatestrings=None):
    """
    Compute the named "GateSet & Dataset" quantity.

    Parameters
    ----------
    qtyname : string
        Name of the quantity to compute.

    gateset : GateSet
        Gate set used to compute the quantity.

    dataset : DataSet
        Data used to compute the quantity.

    gatestrings : list of tuples or GateString objects, optional
        A list of gatestrings used in the computation of certain quantities.
        If None, all the gatestrings in the dataset are used.

    Returns
    -------
    ReportableQty
        The quantity requested, or None if quantity could not be computed.
    """
    ret = compute_gateset_dataset_qtys( [qtyname], gateset, dataset, gatestrings )
    if qtyname is None: return ret
    elif qtyname in ret: return ret[qtyname]
    else: return None

def compute_gateset_dataset_qtys(qtynames, gateset, dataset, gatestrings=None):
    """
    Compute the named "GateSet & Dataset" quantities.

    Parameters
    ----------
    qtynames : list of strings
        Names of the quantities to compute.

    gateset : GateSet
        Gate set used to compute the quantities.

    dataset : DataSet
        Data used to compute the quantities.

    gatestrings : list of tuples or GateString objects, optional
        A list of gatestrings used in the computation of certain quantities.
        If None, all the gatestrings in the dataset are used.

    Returns
    -------
    dict
        Dictionary whose keys are the requested quantity names and values are
        ReportableQty objects.
    """

    #Note: no error bars computed for these quantities yet...

    ret = _OrderedDict()
    possible_qtys = [ ]

    #Quantities computed per gatestring
    per_gatestring_qtys = _OrderedDict() # OLD qtys: [('logl term diff', []), ('score', [])]
    for spl in gateset.get_spam_labels():
        per_gatestring_qtys['prob(%s) diff' % spl] = []
        per_gatestring_qtys['count(%s) diff' % spl] = []
        per_gatestring_qtys['Est prob(%s)' % spl] = []
        per_gatestring_qtys['Est count(%s)' % spl] = []
        per_gatestring_qtys['gatestring chi2(%s)' % spl] = []

    if any( [qtyname in per_gatestring_qtys for qtyname in qtynames ] ):
        if gatestrings is None: gatestrings = list(dataset.keys())
        for gs in gatestrings:
            if gs in dataset: # skip gate strings given that are not in dataset
                dsRow = dataset[gs]
            else: continue

            p = gateset.probs(gs)
            pExp = { }; N = dsRow.total()
            for spamLabel in p:
                p[spamLabel] = _projectToValidProb( p[spamLabel], tol=1e-10 )
                pExp[spamLabel] = _projectToValidProb( dsRow[spamLabel] / N, tol=1e-10 )

            #OLD
            #per_gatestring_qtys['logl term diff'].append(  _tools.logL_term(dsRow, pExp) - _tools.logL_term(dsRow, p)  )
            #per_gatestring_qtys['score'].append(  (_tools.logL_term(dsRow, pExp) - _tools.logL_term(dsRow, p)) / N  )

            for spamLabel in p:
                per_gatestring_qtys['prob(%s) diff' % spamLabel].append( abs(p[spamLabel] - pExp[spamLabel]) )
                per_gatestring_qtys['count(%s) diff' % spamLabel].append( int( round(p[spamLabel] * N) - dsRow[spamLabel]) )
                per_gatestring_qtys['Est prob(%s)' % spamLabel].append( p[spamLabel] )
                per_gatestring_qtys['Est count(%s)' % spamLabel].append( int(round(p[spamLabel] * N)) )
                per_gatestring_qtys['gatestring chi2(%s)' % spamLabel].append( _tools.chi2fn( N, p[spamLabel], pExp[spamLabel], 1e-4 ) )

        for qtyname in qtynames:
            if qtyname in per_gatestring_qtys:
                ret[qtyname] = ReportableQty( per_gatestring_qtys[qtyname] )

    #Quantities which take a single value for a given gateset and dataset
    qty = "logl"; possible_qtys.append(qty)
    if qty in qtynames:
        ret[qty] = ReportableQty( _tools.logl(gateset, dataset) )

    qty = "logl diff"; possible_qtys.append(qty)
    if qty in qtynames:
        ret[qty] = ReportableQty( _tools.logl_max(dataset) - _tools.logl(gateset, dataset) )

    qty = "chi2"; possible_qtys.append(qty)
    if qty in qtynames:
        ret[qty] = ReportableQty( _tools.chi2( dataset, gateset, minProbClipForWeighting=1e-4) )

    #Quantities which take a single value per spamlabel for a given gateset and dataset
    #for spl in gateset.get_spam_labels():
    #    qty = "chi2(%s)" % spl; possible_qtys.append(qty)
    #    if qty in qtynames:
    #        ret[qty] = _tools.chi2( dataset, gateset, minProbClipForWeighting=1e-4)

    if qtynames[0] is None:
        return possible_qtys + list(per_gatestring_qtys.keys())
    return ret


def compute_gateset_gateset_qty(qtyname, gateset1, gateset2,
                                confidenceRegionInfo=None):
    """
    Compute the named "GateSet vs. GateSet" quantity.

    Parameters
    ----------
    qtyname : string
        Name of the quantity to compute.

    gateset1 : GateSet
        First gate set used to compute the quantity.

    gateset2 : GateSet
        Second gate set used to compute the quantity.

    confidenceRegionInfo : ConfidenceRegion, optional
        If not None, specifies a confidence-region used to compute the error bars
        contained in the returned quantity.  If None, then no error bars are
        computed.

    Returns
    -------
    ReportableQty
        The quantity requested, or None if quantity could not be computed.
    """
    ret = compute_gateset_gateset_qtys( [qtyname], gateset1, gateset2, confidenceRegionInfo)
    if qtyname is None: return ret
    elif qtyname in ret: return ret[qtyname]
    else: return None

def compute_gateset_gateset_qtys(qtynames, gateset1, gateset2,
                                 confidenceRegionInfo=None):
    """
    Compute the named "GateSet vs. GateSet" quantities.

    Parameters
    ----------
    qtynames : list of strings
        Names of the quantities to compute.

    gateset1 : GateSet
        First gate set used to compute the quantities.

    gateset2 : GateSet
        Second gate set used to compute the quantities.

    confidenceRegionInfo : ConfidenceRegion, optional
        If not None, specifies a confidence-region used to compute the error bars
        contained in the returned quantities.  If None, then no error bars are
        computed.

    Returns
    -------
    dict
        Dictionary whose keys are the requested quantity names and values are
        ReportableQty objects.
    """
    ret = _OrderedDict()
    possible_qtys = [ ]
    eps = FINITE_DIFF_EPS

    for gateLabel in gateset1.gates:
        if gateLabel not in gateset2.gates:
            raise ValueError("%s gate is missing from second gateset - cannot compare gatesets", gateLabel)
    for gateLabel in gateset2.gates:
        if gateLabel not in gateset1.gates:
            raise ValueError("%s gate is missing from first gateset - cannot compare gatesets", gateLabel)

    mxBasis = gateset1.get_basis_name()
    if mxBasis != gateset2.get_basis_name():
        raise ValueError("Basis mismatch: %s != %s" %
                         (mxBasis, gateset2.get_basis_name()))

    ### per gate quantities
    #############################################
    for gateLabel in gateset1.gates:

        key = '%s fidelity' % gateLabel; possible_qtys.append(key)
        key2 = '%s infidelity' % gateLabel; possible_qtys.append(key)
        if key in qtynames or key2 in qtynames:

            def process_fidelity(gate):
                return _tools.process_fidelity(gate, gateset2.gates[gateLabel], mxBasis)
                  #vary elements of gateset1 (assume gateset2 is fixed)

            #print "DEBUG: fidelity(%s)" % gateLabel
            FQty = _getGateQuantity(process_fidelity, gateset1, gateLabel,
                                    eps, confidenceRegionInfo)

            InFQty = ReportableQty( 1.0-FQty.get_value(), FQty.get_err_bar() )
            if key in qtynames: ret[key] = FQty
            if key2 in qtynames: ret[key2] = InFQty

        key = '%s closest unitary fidelity' % gateLabel; possible_qtys.append(key)
        if key in qtynames:

            #Note: default 'gm' basis
            def closest_unitary_fidelity(gate): # assume vary gateset1, gateset2 fixed
                decomp1 = _tools.decompose_gate_matrix(gate)
                decomp2 = _tools.decompose_gate_matrix(gateset2.gates[gateLabel])

                if decomp1['isUnitary']:
                    closestUGateMx1 = gate
                else: closestUGateMx1 = _alg.find_closest_unitary_gatemx(gate)

                if decomp2['isUnitary']:
                    closestUGateMx2 = gateset2.gates[gateLabel]
                else: closestUGateMx2 = _alg.find_closest_unitary_gatemx(gateset2.gates[gateLabel])

                closeChoi1 = _tools.jamiolkowski_iso(closestUGateMx1)
                closeChoi2 = _tools.jamiolkowski_iso(closestUGateMx2)
                return _tools.fidelity(closeChoi1,closeChoi2)

            ret[key] = _getGateQuantity(closest_unitary_fidelity, gateset1, gateLabel, eps, confidenceRegionInfo)

        key = "%s Frobenius diff" % gateLabel; possible_qtys.append(key)
        if key in qtynames:
            def fro_diff(gate): # assume vary gateset1, gateset2 fixed
                return _tools.frobeniusdist(gate,gateset2.gates[gateLabel])
            #print "DEBUG: frodist(%s)" % gateLabel
            ret[key] = _getGateQuantity(fro_diff, gateset1, gateLabel, eps, confidenceRegionInfo)

        key = "%s Jamiolkowski trace dist" % gateLabel; possible_qtys.append(key)
        if key in qtynames:
            def jt_diff(gate): # assume vary gateset1, gateset2 fixed
                return _tools.jtracedist(gate,gateset2.gates[gateLabel], mxBasis)
            #print "DEBUG: jtdist(%s)" % gateLabel
            ret[key] = _getGateQuantity(jt_diff, gateset1, gateLabel, eps, confidenceRegionInfo)

        key = '%s diamond norm' % gateLabel; possible_qtys.append(key)
        if key in qtynames:

            def half_diamond_norm(gate):
                return 0.5 * _tools.diamonddist(gate, gateset2.gates[gateLabel], mxBasis)
                  #vary elements of gateset1 (assume gateset2 is fixed)

            try:
                ret[key] = _getGateQuantity(half_diamond_norm, gateset1, gateLabel,
                                            eps, confidenceRegionInfo)
            except ImportError: #if failed to import cvxpy (probably b/c it's not installed)
                ret[key] = ReportableQty(_np.nan) # report NAN for diamond norms

        key = '%s angle btwn rotn axes' % gateLabel; possible_qtys.append(key)
        if key in qtynames:

            def angle_btwn_axes(gate): #Note: default 'gm' basis
                decomp = _tools.decompose_gate_matrix(gate)
                decomp2 = _tools.decompose_gate_matrix(gateset2.gates[gateLabel])
                axisOfRotn = decomp.get('axis of rotation',None)
                rotnAngle = decomp.get('pi rotations','X')
                axisOfRotn2 = decomp2.get('axis of rotation',None)
                rotnAngle2 = decomp2.get('pi rotations','X')

                if rotnAngle == 'X' or abs(rotnAngle) < 1e-4 or \
                   rotnAngle2 == 'X' or abs(rotnAngle2) < 1e-4:
                    return _np.nan

                if axisOfRotn is None or axisOfRotn2 is None:
                    return _np.nan

                real_dot =  _np.clip( _np.real(_np.dot(axisOfRotn,axisOfRotn2)), -1.0, 1.0)
                return _np.arccos( abs(real_dot) ) / _np.pi
                  #Note: abs() allows axis to be off by 180 degrees -- if showing *angle* as
                  #      well, must flip sign of angle of rotation if you allow axis to
                  #      "reverse" by 180 degrees.

            ret[key] = _getGateQuantity(angle_btwn_axes, gateset1, gateLabel,
                                    eps, confidenceRegionInfo)

        key = '%s relative logTiG eigenvalues' % gateLabel; possible_qtys.append(key)
        if key in qtynames:
            def rel_eigvals(gate):
                rel_gate = _tools.error_generator(gate, gateset2.gates[gateLabel], "logTiG")
                return _np.linalg.eigvals(rel_gate)
                  #vary elements of gateset1 (assume gateset2 is fixed)

            ret[key] = _getGateQuantity(rel_eigvals, gateset1, gateLabel,
                                        eps, confidenceRegionInfo)

        key = '%s relative logG-logT eigenvalues' % gateLabel; possible_qtys.append(key)
        if key in qtynames:
            def rel_eigvals(gate):
                rel_gate = _tools.error_generator(gate, gateset2.gates[gateLabel], "logG-logT")
                return _np.linalg.eigvals(rel_gate)
                  #vary elements of gateset1 (assume gateset2 is fixed)

            ret[key] = _getGateQuantity(rel_eigvals, gateset1, gateLabel,
                                        eps, confidenceRegionInfo)


    ### per prep vector quantities
    #############################################
    for prepLabel in gateset1.get_prep_labels():

        key = '%s prep state fidelity' % prepLabel; possible_qtys.append(key)
        key2 = '%s prep state infidelity' % prepLabel; possible_qtys.append(key)
        if key in qtynames or key2 in qtynames:

            def fidelity(vec):
                rhoMx1 = _tools.vec_to_stdmx(vec, mxBasis)
                rhoMx2 = _tools.vec_to_stdmx(gateset2.preps[prepLabel], mxBasis)
                return _tools.fidelity(rhoMx1, rhoMx2)
                  #vary elements of gateset1 (assume gateset2 is fixed)

            FQty = _getPrepQuantity(fidelity, gateset1, prepLabel,
                                    eps, confidenceRegionInfo)

            InFQty = ReportableQty( 1.0-FQty.get_value(), FQty.get_err_bar() )
            if key in qtynames: ret[key] = FQty
            if key2 in qtynames: ret[key2] = InFQty

        key = "%s prep trace dist" % prepLabel; possible_qtys.append(key)
        if key in qtynames:
            def tr_diff(vec): # assume vary gateset1, gateset2 fixed
                rhoMx1 = _tools.vec_to_stdmx(vec, mxBasis)
                rhoMx2 = _tools.vec_to_stdmx(gateset2.preps[prepLabel], mxBasis)
                return _tools.tracedist(rhoMx1, rhoMx2)
            ret[key] = _getPrepQuantity(tr_diff, gateset1, prepLabel,
                                        eps, confidenceRegionInfo)


    ### per effect vector quantities
    #############################################
    for effectLabel in gateset1.get_effect_labels():

        key = '%s effect state fidelity' % effectLabel; possible_qtys.append(key)
        key2 = '%s effect state infidelity' % effectLabel; possible_qtys.append(key)
        if key in qtynames or key2 in qtynames:

            def fidelity(vec):
                EMx1 = _tools.vec_to_stdmx(vec, mxBasis)
                EMx2 = _tools.vec_to_stdmx(gateset2.effects[effectLabel], mxBasis)
                return _tools.fidelity(EMx1,EMx2)
                  #vary elements of gateset1 (assume gateset2 is fixed)

            FQty = _getEffectQuantity(fidelity, gateset1, effectLabel,
                                      eps, confidenceRegionInfo)

            InFQty = ReportableQty( 1.0-FQty.get_value(), FQty.get_err_bar() )
            if key in qtynames: ret[key] = FQty
            if key2 in qtynames: ret[key2] = InFQty

        key = "%s effect trace dist" % effectLabel; possible_qtys.append(key)
        if key in qtynames:
            def tr_diff(vec): # assume vary gateset1, gateset2 fixed
                EMx1 = _tools.vec_to_stdmx(vec, mxBasis)
                EMx2 = _tools.vec_to_stdmx(gateset2.effects[effectLabel], mxBasis)
                return _tools.tracedist(EMx1, EMx2)
            ret[key] = _getEffectQuantity(tr_diff, gateset1, effectLabel,
                                          eps, confidenceRegionInfo)


    ###  per gateset quantities
    #############################################
    key = "Gateset Frobenius diff"; possible_qtys.append(key)
    if key in qtynames: ret[key] = ReportableQty( gateset1.frobeniusdist(gateset2) )

    key = "Max Jamiolkowski trace dist"; possible_qtys.append(key)
    if key in qtynames: ret[key] = ReportableQty(
        max( [ _tools.jtracedist(gateset1.gates[l],gateset2.gates[l])
               for l in gateset1.gates ] ) )


    #Special case: when qtyname is None then return a list of all possible names that can be computed
    if qtynames[0] is None:
        return possible_qtys
    return ret



from __future__ import division, print_function, absolute_import, unicode_literals
#*****************************************************************
#    pyGSTi 0.9:  Copyright 2015 Sandia Corporation
#    This Software is released under the GPL license detailed
#    in the file "license.txt" in the top-level pyGSTi directory
#*****************************************************************
"""
Variables for working with the a gate set containing Idle, X(pi/2) and Y(pi/2) gates.
"""

from . import gatestringconstruction as _strc
from . import gatesetconstruction as _setc

description = "Idle, X(pi/2), and Y(pi/2) gates"

gates = ['Gi','Gx','Gy']
fiducials = _strc.gatestring_list( [ (), ('Gx',), ('Gy',), ('Gx','Gx'),
                                     ('Gx','Gx','Gx'), ('Gy','Gy','Gy') ] ) # for 1Q MUB
germs = _strc.gatestring_list( [('Gx',), ('Gy',), ('Gi',), ('Gx', 'Gy'),
                                ('Gx', 'Gy', 'Gi'), ('Gx', 'Gi', 'Gy'), ('Gx', 'Gi', 'Gi'), ('Gy', 'Gi', 'Gi'),
                                  ('Gx', 'Gx', 'Gi', 'Gy'), ('Gx', 'Gy', 'Gy', 'Gi'),
                                  ('Gx', 'Gx', 'Gy', 'Gx', 'Gy', 'Gy')] )

#Construct a target gateset: Identity, X(pi/2), Y(pi/2)
gs_target = _setc.build_gateset([2],[('Q0',)], ['Gi','Gx','Gy'],
                                [ "I(Q0)","X(pi/2,Q0)", "Y(pi/2,Q0)"],
                                 prepLabels=["rho0"], prepExpressions=["0"],
                                 effectLabels=["E0"], effectExpressions=["1"],
                                 spamdefs={'plus': ('rho0','E0'), 'minus': ('rho0','remainder') } )



from __future__ import division, print_function, absolute_import, unicode_literals
#*****************************************************************
#    pyGSTi 0.9:  Copyright 2015 Sandia Corporation
#    This Software is released under the GPL license detailed
#    in the file "license.txt" in the top-level pyGSTi directory
#*****************************************************************
"""
Variables for working with the a gate set containing Idle, Z(pi/2) and rot(X=pi/2, Y=sqrt(3)/2) gates.
"""

from . import gatestringconstruction as _strc
from . import gatesetconstruction as _setc


gs_target = _setc.build_gateset([2],[('Q0',)], ['Gz','Gn'],
                                [ "Z(pi/2,Q0)", "N(pi/2, sqrt(3)/2, 0, -0.5, Q0)"],
                                prepLabels=["rho0"], prepExpressions=["0"],
                                effectLabels=["E0"], effectExpressions=["1"],
                                spamdefs={'plus': ('rho0','E0'), 'minus': ('rho0','remainder') } )

prepFiducials = _strc.gatestring_list([(),
                                       ('Gn',),
                                       ('Gn','Gn'),
                                       ('Gn','Gz','Gn'),
                                       ('Gn','Gn','Gn',),
                                       ('Gn','Gz','Gn','Gn','Gn')]) # for 1Q MUB

measFiducials = _strc.gatestring_list([(),
                                       ('Gn',),
                                       ('Gn','Gn'),
                                       ('Gn','Gz','Gn'),
                                       ('Gn','Gn','Gn',),
                                       ('Gn','Gn','Gn','Gz','Gn')]) # for 1Q MUB

germs = _strc.gatestring_list([ ('Gz',),
                                ('Gn',),
                                ('Gn','Gn','Gz','Gn','Gz'),
                                ('Gn','Gz','Gn','Gz','Gz'),
                                ('Gn','Gz','Gn','Gn','Gz','Gz'),
                                ('Gn','Gn','Gz','Gn','Gz','Gz'),
                                ('Gn','Gn','Gn','Gz','Gz','Gz') ])



from __future__ import division, print_function, absolute_import, unicode_literals
#*****************************************************************
#    pyGSTi 0.9:  Copyright 2015 Sandia Corporation
#    This Software is released under the GPL license detailed
#    in the file "license.txt" in the top-level pyGSTi directory
#*****************************************************************
""" Gate string list creation functions using repeated-germs limited by a max-length."""

import numpy.random as _rndm
import itertools as _itertools
from ..tools import listtools as _lt
from . import gatestringconstruction as _gsc
from . import spamspecconstruction as _ssc


def make_lsgst_lists(gateLabels, prepStrs, effectStrs, germList, maxLengthList,
                     fidPairs=None, truncScheme="whole germ powers", nest=True,
                     keepFraction=1, keepSeed=None):
    """
    Create a set of gate string lists for LSGST based on germs and max-lengths.

    Constructs a series (a list) of gate string lists used by long-sequence GST
    (LSGST) algorithms.  If maxLengthList[0] == 0 then the starting list is the
    list of LGST strings, otherwise the starting list is empty.  For each
    nonzero element of maxLengthList, call it L, a list of gate strings is
    created with the form:

    Case: truncScheme == 'whole germ powers':
      prepStr + pygsti.construction.repeat_with_max_length(germ,L) + effectStr

    Case: truncScheme == 'truncated germ powers':
      prepStr + pygsti.construction.repeat_and_truncate(germ,L) + effectStr

    Case: truncScheme == 'length as exponent':
      prepStr + germ^L + effectStr

    If nest == True, the above list is iteratively *added* (w/duplicates
    removed) to the current list of gate strings to form a final list for the
    given L.  This results in successively larger lists, each of which
    contains all the elements of previous-L lists.  If nest == False then
    the above list *is* the final list for the given L.

    Parameters
    ----------
    gateLabels : list or tuple
        List of gate labels to determine needed LGST strings.  Only relevant
        when maxLengthList[0] == 0.

    prepStrs : list of GateStrings
        List of the preparation fiducial gate strings, which follow state
        preparation.

    effectStrs : list of GateStrings
        List of the measurement fiducial gate strings, which precede
        measurement.

    germList : list of GateStrings
        List of the germ gate strings.

    maxLengthList : list of ints
        List of maximum lengths.  If maxLengthList[0] == 0 this results in
        special behavior where LGST strings are included as the first
        returned list.

    fidPairs : list of 2-tuples or dict, optional
        Specifies a subset of all fiducial string pairs (prepStr, effectStr)
        to be used in the gate string lists.  If a list, each element of 
        fidPairs is a (iPrepStr, iEffectStr) 2-tuple of integers, each 
        indexing a string within prepStrs and effectStrs, respectively, so 
        that prepStr = prepStrs[iPrepStr] and effectStr = 
        effectStrs[iEffectStr].  If a dictionary, keys are germs (elements
        of germList) and values are lists of 2-tuples specifying the pairs
        to use for that germ.

    truncScheme : str, optional
        Truncation scheme used to interpret what the list of maximum lengths
        means. If unsure, leave as default. Allowed values are:

        - 'whole germ powers' -- germs are repeated an integer number of
          times such that the length is less than or equal to the max.
        - 'truncated germ powers' -- repeated germ string is truncated
          to be exactly equal to the max (partial germ at end is ok).
        - 'length as exponent' -- max. length is instead interpreted
          as the germ exponent (the number of germ repetitions).

    nest : boolean, optional
        If Frue, the returned gate string lists are "nested", meaning
        that each successive list of gate strings contains all the gate
        strings found in previous lists (and usually some additional
        new ones).  If False, then the returned string list for maximum
        length == L contains *only* those gate strings specified in the
        description above, and *not* those for previous values of L.

    keepFraction : float, optional
        The fraction of fiducial pairs selected for each germ-power base
        string.  The default includes all fiducial pairs.  Note that
        for each germ-power the selected pairs are *different* random
        sets of all possible pairs (unlike fidPairs, which specifies the
        *same* fiducial pairs for *all* same-germ base strings).  If
        fidPairs is used in conjuction with keepFraction, the pairs
        specified by fidPairs are always selected, and any additional
        pairs are randomly selected.

    keepSeed : int, optional
        The seed used for random fiducial pair selection (only relevant
        when keepFraction < 1).

    Returns
    -------
    list of (lists of GateStrings)
        The i-th list corresponds to a gate string list containing repeated
        germs limited to length maxLengthList[i].  If nest == True, then
        repeated germs limited to previous max-lengths are also included.
        Note that a "0" maximum-length corresponds to the LGST strings.
    """
    lgstStrings = _gsc.list_lgst_gatestrings( _ssc.build_spam_specs(prepStrs = prepStrs, effectStrs = effectStrs),
                                              gateLabels)
    lsgst_list = _gsc.gatestring_list([ () ]) #running list of all strings so far

    if keepFraction < 1.0:
        rndm = _rndm.RandomState(keepSeed) # ok if seed is None
        nPairs = len(prepStrs)*len(effectStrs)
        nPairsToKeep = int(round(float(keepFraction) * nPairs))
    else: rndm = None

    if isinstance(fidPairs, dict) or hasattr(fidPairs, "keys"):
        fiducialPairs = { germ: [ (prepStrs[i],effectStrs[j]) 
                                  for (i,j) in fidPairs[germ] ]
                          for germ in germList }
        fidPairDict = fidPairs
    else:
        if fidPairs is not None:   #assume fidPairs is a list
            fidPairDict = { germ:fidPairs for germ in germList }
            lst = [ (prepStrs[i],effectStrs[j]) for (i,j) in fidPairs ]
        else:
            fidPairDict = None
            lst = list(_itertools.product(prepStrs, effectStrs))
        fiducialPairs = { germ:lst for germ in germList }

        
    if maxLengthList[0] == 0:
        lsgst_listOfLists = [ lgstStrings ]
        maxLengthList = maxLengthList[1:]
    else: lsgst_listOfLists = [ ]

    Rfn = _getTruncFunction(truncScheme)

    for maxLen in maxLengthList:

        lst = []
        for germ in germList:

            if rndm is None:
                fiducialPairsThisIter = fiducialPairs[germ]

            elif fidPairDict is not None:
                pair_indx_tups = fidPairDict[germ]
                remainingPairs = [ (prepStrs[i],effectStrs[j])
                                   for i in range(len(prepStrs))
                                   for j in range(len(effectStrs))
                                   if (i,j) not in pair_indx_tups ]
                nPairsRemaining = len(remainingPairs)
                nPairsToChoose = nPairsToKeep-len(pair_indx_tups)
                nPairsToChoose = max(0,min(nPairsToChoose,nPairsRemaining))
                assert(0 <= nPairsToChoose <= nPairsRemaining)
                # FUTURE: issue warnings when clipping nPairsToChoose?

                fiducialPairsThisIter = fiducialPairs[germ] + \
                    [ remainingPairs[k] for k in
                      sorted(rndm.choice(nPairsRemaining,nPairsToChoose,
                                         replace=False))]

            else: # rndm is not None and fidPairDict is None
                assert(nPairsToKeep <= nPairs) # keepFraction must be <= 1.0
                fiducialPairsThisIter = \
                    [ fiducialPairs[germ][k] for k in
                      sorted(rndm.choice(nPairs,nPairsToKeep,replace=False))]

            lst += _gsc.create_gatestring_list("f[0]+R(germ,N)+f[1]",
                                              f=fiducialPairsThisIter,
                                              germ=germ, N=maxLen,
                                              R=Rfn, order=('f',))
        if nest:
            lsgst_list += lst #add new strings to running list
            lsgst_listOfLists.append( _lt.remove_duplicates(lgstStrings + lsgst_list) )
        else:
            lsgst_listOfLists.append( _lt.remove_duplicates(lst) )

    #print "%d LSGST sets w/lengths" % len(lsgst_listOfLists),map(len,lsgst_listOfLists)
    return lsgst_listOfLists


def make_lsgst_experiment_list(gateLabels, prepStrs, effectStrs, germList,
                               maxLengthList, fidPairs=None,
                               truncScheme="whole germ powers", keepFraction=1,
                               keepSeed=None):
    """
    Create a list of all the gate strings (i.e. the experiments) required for
    long-sequence GST (LSGST) algorithms.

    Returns a single list containing, without duplicates, all the gate
    strings required throughout all the iterations of LSGST given by
    maxLengthList.  Thus, the returned list is equivalently the list of
    the experiments required to run LSGST using the supplied parameters,
    and so commonly used when construting data set templates or simulated
    data sets.  The breakdown of which gate strings are used for which
    iteration(s) of LSGST is given by make_lsgst_lists(...).

    Parameters
    ----------
    gateLabels : list or tuple
        List of gate labels to determine needed LGST strings.  Only relevant
        when maxLengthList[0] == 0.

    prepStrs : list of GateStrings
        List of the preparation fiducial gate strings, which follow state
        preparation.

    effectStrs : list of GateStrings
        List of the measurement fiducial gate strings, which precede
        measurement.

    germList : list of GateStrings
        List of the germ gate strings.

    maxLengthList : list of ints
        List of maximum lengths.  If maxLengthList[0] == 0 this results in
        special behavior where LGST strings are included as the first
        returned list.

    fidPairs : list of 2-tuples, optional
        Specifies a subset of all fiducial string pairs (prepStr, effectStr)
        to be used in the gate string lists.  If a list, each element of 
        fidPairs is a (iPrepStr, iEffectStr) 2-tuple of integers, each 
        indexing a string within prepStrs and effectStrs, respectively, so 
        that prepStr = prepStrs[iPrepStr] and effectStr = 
        effectStrs[iEffectStr].  If a dictionary, keys are germs (elements
        of germList) and values are lists of 2-tuples specifying the pairs
        to use for that germ.

    truncScheme : str, optional
        Truncation scheme used to interpret what the list of maximum lengths
        means. If unsure, leave as default. Allowed values are:

        - 'whole germ powers' -- germs are repeated an integer number of
          times such that the length is less than or equal to the max.
        - 'truncated germ powers' -- repeated germ string is truncated
          to be exactly equal to the max (partial germ at end is ok).
        - 'length as exponent' -- max. length is instead interpreted
          as the germ exponent (the number of germ repetitions).

    keepFraction : float, optional
        The fraction of fiducial pairs selected for each germ-power base
        string.  The default includes all fiducial pairs.  Note that
        for each germ-power the selected pairs are *different* random
        sets of all possible pairs (unlike fidPairs, which specifies the
        *same* fiduicial pairs for *all* same-germ base strings).  If
        fidPairs is used in conjuction with keepFraction, the pairs
        specified by fidPairs are always selected, and any additional
        pairs are randomly selected.

    keepSeed : int, optional
        The seed used for random fiducial pair selection (only relevant
        when keepFraction < 1).


    Returns
    -------
    list of GateStrings
    """
    nest = True # => the final list contains all of the strings
    return make_lsgst_lists(gateLabels, prepStrs, effectStrs, germList,
                            maxLengthList, fidPairs, truncScheme, nest,
                            keepFraction, keepSeed)[-1]



def make_elgst_lists(gateLabels, germList, maxLengthList,
                     truncScheme="whole germ powers", nest=True):
    """
    Create a set of gate string lists for eLGST based on germs and max-lengths

    Constructs a series (a list) of gate string lists used by the extended LGST
    (eLGST) algorithm.  If maxLengthList[0] == 0 then the starting list is the
    list of length-1 gate label strings, otherwise the starting list is empty.
    For each nonzero element of maxLengthList, call it L, a list of gate strings is
    created with the form:

    Case: truncScheme == 'whole germ powers':
      pygsti.construction.repeat_with_max_length(germ,L)

    Case: truncScheme == 'truncated germ powers':
      pygsti.construction.repeat_and_truncate(germ,L)

    Case: truncScheme == 'length as exponent':
      germ^L

    If nest == True, the above list is iteratively *added* (w/duplicates
    removed) to the current list of gate strings to form a final list for the
    given L.  This results in successively larger lists, each of which
    contains all the elements of previous-L lists.  If nest == False then
    the above list *is* the final list for the given L.

    Parameters
    ----------
    gateLabels : list or tuple
        List of gate labels. Only relevant when maxLengthList[0] == 0.

    germList : list of GateStrings
        List of the germ gate strings.

    maxLengthList : list of ints
        List of the maximum lengths.  If maxLengthList[0] == 0 this results in
        special behavior where the length-1 gate label strings are included as
        the first returned list.

    truncScheme : str, optional
        Truncation scheme used to interpret what the list of maximum lengths
        means.  If unsure, leave as default. Allowed values are:

        - 'whole germ powers' -- germs are repeated an integer number of
          times such that the length is less than or equal to the max.
        - 'truncated germ powers' -- repeated germ string is truncated
          to be exactly equal to the max (partial germ at end is ok).
        - 'length as exponent' -- max. length is instead interpreted
          as the germ exponent (the number of germ repetitions).

    nest : boolean, optional
        If Frue, the returned gate string lists are "nested", meaning
        that each successive list of gate strings contains all the gate
        strings found in previous lists (and usually some additional
        new ones).  If False, then the returned string list for maximum
        length == L contains *only* those gate strings specified in the
        description above, and *not* those for previous values of L.


    Returns
    -------
    list of (lists of GateStrings)
        The i-th list corresponds to a gate string list containing repeated
        germs limited to length maxLengthList[i].  If nest == True, then
        repeated germs limited to previous max-lengths are also included.
        Note that a "0" maximum-length corresponds to the gate
        label strings.
    """
    singleGates = _gsc.gatestring_list([(g,) for g in gateLabels])
    elgst_list = _gsc.gatestring_list([ () ])  #running list of all strings so far

    if maxLengthList[0] == 0:
        elgst_listOfLists = [ singleGates ]
        maxLengthList = maxLengthList[1:]
    else: elgst_listOfLists = [ ]

    Rfn = _getTruncFunction(truncScheme)

    for maxLen in maxLengthList:
        lst = _gsc.create_gatestring_list("R(germ,N)", germ=germList, N=maxLen, R=Rfn)
        if nest:
            elgst_list += lst #add new strings to running list
            elgst_listOfLists.append( _lt.remove_duplicates(singleGates + elgst_list) )
        else:
            elgst_listOfLists.append( _lt.remove_duplicates(lst) )

    #print "%d eLGST sets w/lengths" % len(elgst_listOfLists),map(len,elgst_listOfLists)
    return elgst_listOfLists


def make_elgst_experiment_list(gateLabels, germList, maxLengthList,
                               truncScheme="whole germ powers"):
    """
    Create a list of all the gate strings (i.e. the experiments) required for
    the extended LGST (eLGST) algorithm.

    Returns a single list containing, without duplicates, all the gate
    strings required throughout all the iterations of eLGST given by
    maxLengthList.  Thus, the returned list is equivalently the list of
    the experiments required to run eLGST using the supplied parameters,
    and so commonly used when construting data set templates or simulated
    data sets.  The breakdown of which gate strings are used for which
    iteration(s) of eLGST is given by make_elgst_lists(...).

    Parameters
    ----------
    gateLabels : list or tuple
        List of gate labels. Only relevant when maxLengthList[0] == 0.

    germList : list of GateStrings
        List of the germ gate strings.

    maxLengthList : list of ints
        List of the maximum lengths.  If maxLengthList[0] == 0 this results in
        special behavior where the length-1 gate label strings are included as
        the first returned list.

    truncScheme : str, optional
        Truncation scheme used to interpret what the list of maximum lengths
        means. If unsure, leave as default. Allowed values are:

        - 'whole germ powers' -- germs are repeated an integer number of
          times such that the length is less than or equal to the max.
        - 'truncated germ powers' -- repeated germ string is truncated
          to be exactly equal to the max (partial germ at end is ok).
        - 'length as exponent' -- max. length is instead interpreted
          as the germ exponent (the number of germ repetitions).

    Returns
    -------
    list of GateStrings
    """

    #When nest == True the final list contains all of the strings
    return make_elgst_lists(gateLabels, germList,
                            maxLengthList, truncScheme, nest=True)[-1]



def _getTruncFunction(truncScheme):
    if truncScheme == "whole germ powers":
        Rfn = _gsc.repeat_with_max_length
    elif truncScheme == "truncated germ powers":
        Rfn = _gsc.repeat_and_truncate
    elif truncScheme == "length as exponent":
        def Rfn(germ,N): return germ*N
    else:
        raise ValueError("Invalid truncation scheme: %s" % truncScheme)
    return Rfn
